/*	vote.c

	AFAPI voting and scheduling function definitions for
	TTL_PAPERS.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"

/*	p_first()...

	Return the IPROC of the first lowest-numbered PE with a yes
	vote.  For SHMAPERS, this is very similar to p_any().
*/

int
p_first(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 1... */
	p = p_shm_all;
	do {
		if (p->datum[ which ].u_1u) return(p - p_shm_all);
	} while (++p < q);

	return(NPROC);
}

/*	p_count()...

	Count how many voted yes.
*/

inline int
p_count(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register int count = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for 1 values and count them... */
	p = p_shm_all;
	do {
		count += p->datum[ which ].u_1u;
	} while (++p < q);

	return(count);
}

/*	p_quantify()...

	Quantify the number of PEs voting yes; the return is 0, 1, 2
	(really >1), or NPROC.  The value is allowed to distinguish
	more than these counts, and actually returns a precise count
	in some implementations.  It turns out that SHMAPERS can't do
	any better than always using p_count()....
*/

int
p_quantify(register p_1u f)
{
	return(p_count(f));
}

/*	p_vote()...

	Builds a vote barrier bit vector much as p_waitbar().
*/

inline p_bar
p_vote(register int pe)
{
	/* Collect the array of votes and construct bit vector... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register p_bar result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_32 = pe;
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	p = p_shm_all;
	do {
		--q;
		result = ((result << 1) |
			  (q->datum[ which ].u_32 == IPROC));
	} while (q > p);

	return(result);
}

/*	p_voteCount()...
*/

int
p_voteCount(register int pe)
{
	/* Collect the array of votes and count matches... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register int result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_32 = pe;
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	p = p_shm_all;
	do {
		--q;
		result += (q->datum[ which ].u_32 == IPROC);
	} while (q > p);

	return(result);
}

/*	p_match()...
*/

#define	P_DECLARE(suffix) \
inline p_bar \
p_match##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register p_bar result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Construct bit vector... */ \
	p = p_shm_all; \
	do { \
		--q; \
		result = ((result << 1) | \
			  (q->datum[ which ].u_##suffix == datum)); \
	} while (q > p); \
 \
	return(result); \
}

P_BETA

#undef	P_DECLARE

/*	p_matchCount()...
*/

#define	P_DECLARE(suffix) \
int \
p_matchCount##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register int result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Count matches... */ \
	p = p_shm_all; \
	do { \
		--q; \
		result += (q->datum[ which ].u_##suffix == datum); \
	} while (q > p); \
 \
	return(result); \
}

P_BETA

#undef	P_DECLARE

/*	p_rank()...
*/

#define	P_DECLARE(suffix) \
int \
p_rank##suffix(register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
	register int result = 0; \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Count things ranking below us... */ \
	p = p_shm_all; \
	do { \
		register p_##suffix t = q->datum[ which ].u_##suffix; \
		result += ((t < datum) || \
			   ((t == datum) && (p < p_shm_me))); \
	} while (++p < q); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE
