/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	sockface.str.c,v $
 * Revision 6.0  96/02/29  13:07:14  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/13  16:44:01  nevin
 * Made calculation of kernel address length more robust.
 * 
 * Revision 5.2.1.1  1996/02/07 01:49:51  gdburns
 * Add stream, UNIX domain, socket functions.
 *
 * Revision 5.2  94/08/22  13:35:11  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  10:35:18  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/04/22  12:08:25  gdburns
 * Ohio Release
 * 
 *	Function:	- socket interface functions
 *			- STREAM version
 *
 *			- getsrvstrsock(): create server-side socket
 *			- getcltstrsock(): create client-side socket
 *			- accepttmout(): accept connection with timeout
 */

#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/un.h>
#include <unistd.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <portable.h>
#include <typical.h>

/*
 * external functions
 */
extern struct hostent	*gethostbyname();

/*
 *	getsrvstrsock
 *
 *	Function:	- create a server-side socket
 *			- Internet Domain, Stream connection
 *			- if pport points to a non-zero positive port
 *			  number, that port number is used otherwise a
 *			  port number is dynamically allocated
 *			- if pport is non-NULL, the server port number
 *			  is returned in it
 *	Accepts:	- backlog queue size
 *			- ptr to port number (or NULL)
 *	Returns:	- socket descriptor or ERROR
 */
int
getsrvstrsock(backlog, pport)

int			backlog;
int			*pport;

{
	int		sockd;		/* socket descriptor */
	int		length;		/* length of socket name */
	struct sockaddr_in
			srvaddr;	/* server address info */
/*
 * Create server socket.
 */
	if ((sockd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		return(ERROR);
	}
/*
 * Bind the socket.  If a port # is given use it otherwise bind allocates one.
 */
	memset((char *) &srvaddr, 0, sizeof(srvaddr));
	srvaddr.sin_family = AF_INET;
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	srvaddr.sin_port = (pport && (*pport > 0)) ?
			htons((unsigned short) *pport) : htons(0);

	if (bind(sockd, (struct sockaddr *) &srvaddr, sizeof(srvaddr))) {
		close(sockd);
		return(ERROR);
	}
/*
 * If bind allocated a port # dynamically and pport is non-NULL,
 * get the port # allocated and return it in pport.
 */
	if (pport && (*pport <= 0)) {
		length = sizeof(srvaddr);
		if (getsockname(sockd, (struct sockaddr *) &srvaddr, &length)) {
			close(sockd);
			return(ERROR);
		}

		*pport = (int) ntohs(srvaddr.sin_port);
	}
/*
 * Create backlog queue.
 */
	if (listen(sockd, backlog)) {
		close(sockd);
		return(ERROR);
	}

	return(sockd);
}

/*
 *	getcltstrsock
 *
 *	Function:	- create a client-side socket
 *			- Internet Domain, Stream connection
 *	Accepts:	- server host address (4 bytes, in network byte order)
 *			- server port number
 *	Returns:	- socket descriptor or ERROR
 */
int
getcltstrsock(hostaddr, portnum)

unsigned char		*hostaddr;
int			portnum;

{
	int		sockd;		/* socket descriptor */
	int		ret;		/* connect() return value */
	struct sockaddr_in
			srvaddr;	/* server address info */
/*
 * Loop until successful at creating and connecting
 * a socket with no interruption.
 */
	for (;;) {
/*
 * Create client socket.
 */
		if ((sockd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			return(ERROR);
		}
/*
 * Setup the server Internet address structure.
 */
		memset((char *) &srvaddr, 0, sizeof(srvaddr));
		srvaddr.sin_family = AF_INET;
		srvaddr.sin_port = htons((unsigned short) portnum);
		memcpy((char *) &srvaddr.sin_addr, (char *) hostaddr, 4);
/*
 * Connect client socket to server socket.
 */
		ret = connect(sockd,
			(struct sockaddr *) &srvaddr, sizeof(srvaddr));

		if (ret == 0) {
			return(sockd);
		} else if (errno != EINTR) {
			close(sockd);
			return(ERROR);
		} else {
			close(sockd);
		}
	}
}

/*
 *	accepttmout
 *
 *	Function:	- accept a connection with a timeout
 *			- negative timeout values block indefinitely
 *	Accepts:	- socket
 *			- timeout value (seconds)
 *	Returns:	- connection socket or ERROR
 */
int
accepttmout(sock, timeout)

int			sock;
int			timeout;

{
	int		connsock;	/* connection socket */
	int		ret;		/* select return value */
	fd_set		readmask;	/* select read-mask */
	struct timeval	tmout;		/* timeout argument */

	if (timeout >= 0) {
/*
 * Set up the timeout argument and block on a select call.
 */
		tmout.tv_sec = timeout;
		tmout.tv_usec = 0;

		FD_ZERO(&readmask);
		FD_SET(sock, &readmask);

		while (((ret = select(sock + 1, &readmask, (fd_set *) 0,
			(fd_set *) 0, &tmout)) < 0) && (errno == EINTR));

		if (ret <= 0) {
			errno = (ret == 0) ? ETIMEDOUT : errno;
			return(ERROR);
		}
	}
/*
 * Do the accept call.
 */
	while (((connsock = accept(sock, (struct sockaddr *) 0,
			(int *) 0)) < 0) && (errno == EINTR));

	return(connsock);
}

/*
 *	_stm_unix_econnect
 *
 *	Function:	- connects stream, UNIX, client socket
 *			- retries after certain error conditions
 *	Accepts:	- UNIX address (filename)
 *	Returns:	- socket or ERROR
 */
int
_stm_unix_econnect(addr)

char			*addr;

{
	int		sd;
	int		r;
	struct sockaddr_un
			server_un;		/* server address */
/*
 * Set up address.
 */
	server_un.sun_family = AF_UNIX;
	strcpy((char *) server_un.sun_path, addr);
/*
 * Create the socket.
 */
	do {
		if ((sd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
				return(LAMERROR);
/*
 * Connect to kernel.
 */
		r = connect(sd, (struct sockaddr *) &server_un,
				sizeof(server_un) - sizeof(server_un.sun_path)
                                + strlen(addr));

		if (r < 0) {
			if ((errno != ECONNREFUSED) && (errno != EINTR) &&
					(errno != ETIMEDOUT)) {
				close(sd);
				return(LAMERROR);
			}

			if (close(sd)) return(LAMERROR);
		}
	} while (r < 0);

	return(sd);
}
