/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD
 *
 *	$Log:	mpirun.c,v $
 * Revision 6.0.1.2  96/06/13  11:22:44  nevin
 * Rename setmode to setopmode to avoid FreeBSD name clash
 * 
 * Revision 6.0.1.1  96/06/10  15:50:46  nevin
 * Return status of first process in error
 * 
 * Revision 6.0  96/02/29  13:24:45  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.3  96/02/22  01:42:25  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.2  94/09/02  12:14:02  vaigl-j
 * Don't modify pointers you intend to use later.
 * 
 * Revision 5.2.1.1  94/09/01  16:50:26  vaigl-j
 * Free realloc'd pointers, not the originally malloc'd ones.
 * 
 * Revision 5.2  94/08/22  13:43:19  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.4  94/08/02  15:46:37  raja
 * Check and cache -x option early on.
 * 
 * Revision 5.1.1.3  94/07/27  18:22:10  raja
 * Add undocumented -x option for xmpi.
 * 
 * Revision 5.1.1.2  94/05/23  12:24:26  vaigl-j
 * Fix schema parsing bug, and don't free 'fullpath' until all copies loaded.
 * 
 * Revision 5.1.1.1  94/05/19  16:35:15  gdburns
 * Place stdlib.h with system headers.
 * 
 * Revision 5.1  94/05/18  10:39:27  gdburns
 * Ohio Release
 * 
 *	Function:	- start an MPI application
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <all_list.h>
#include <all_opt.h>
#include <app_schema.h>
#include <args.h>
#include <events.h>
#include <freq.h>
#include <kio.h>
#include <mpisys.h>
#include <mpitrace.h>
#include <net.h>
#include <portable.h>
#include <priority.h>
#include <preq.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern struct kio_t	_kio;			/* kernel I/O block */
extern struct fclient	_ufd[FUMAX];		/* user file desc. */

/*
 * local functions
 */
static int		pwait();
static int		sendproctbl();
static int		setstdio();
static void		help();
static void		noderanges();
static void		setopmode();

/*
 * local variables
 */
static int		fd[3] = {0, 1, 2};	/* stdio file desc. */
static int		fl_fdesc;		/* don't open stdio */
static int		fl_nprocs;		/* # procs specified */
static int		fl_verbose;		/* verbose mode */
static int		fl_wait;		/* wait for completion */
static int4		nprocs;			/* # procs */
static int4		rtf;			/* runtime flags */
static char		*infile;		/* input filename */
static char		*outfile;		/* output filename */
static char		fmtbuf[512];		/* formatting buffer */
static char		smallbuf[32];		/* small extra buffer */
static OPT		*ad;			/* argument descriptor */

static char		*syntaxerr =
	"mpirun: application schema syntax error, line %d\n";

static char		*usage =
	"mpirun [options] [<app>] [<prog> [<nodes>] [-- <args>]]\n";

int
main(argc, argv)

int			argc;
char			**argv;

{
	int4		nodeid;			/* target node */
	int4		index;			/* for getnode() */
	int4		nflags;			/* for getnode() */
	int		lineno;			/* line number */
	int		status;			/* return status */
	char		*appschema;		/* application schema */
	LIST		*appd;			/* application desc. */
	int		argtailc;		/* argument tail count */
	char		**argtailv;		/* argument tail vector */
/*
 * Parse the command line.
 */
        ad = ao_init();
        if (ad == 0) {
		perror("mpirun (ao_init)");
		exit(errno);
	}
	ao_setopt1(ad, "fhtvwDO", 0, 0, 0);
	ao_setopt(ad, "c", 0, 1, AOINT);
	ao_setopt(ad, "c2c", 0, 0, 0);
	ao_setopt1(ad, "ios", 0, 1, 0);
	ao_setopt(ad, "nger", 0, 0, 0);
	ao_setopt(ad, "toff", 0, 0, 0);
	ao_setopt(ad, "ton", 0, 0, 0);

	if (ao_parse(ad, &argc, argv))  {
		fprintf(stderr, usage);
		exit(errno);
	}
/*
 * Check for help request.
 */
	if (ao_taken(ad, "h")) {
		help();
		exit(0);
	}
/*
 * Set the operating mode.
 */
	setopmode();
/*
 * Attach to kernel.
 */
	if (kinit(PRCMD)) {
		terror("mpirun (kinit)");
		exit(errno);
	}

	if (nid_parse(&argc, argv) || (errno = (argc > 1) ? 0 : EUSAGE)) {
		fprintf(stderr, usage);
		kexit(errno);
	}
/*
 * Set stdio for the remote filesystem.
 */
	if (setstdio()) lamfail("mpirun");
/*
 * Application schema or command line?
 */
	nid_get(&index, &nodeid, &nflags);

	if ((index < 0) && !fl_nprocs) {
/*
 * Parse the app. schema.
 */
		if ((argc != 2) || ao_taken(ad, "s")) {
			fprintf(stderr, usage);
			kexit(EUSAGE);
		}

		appschema = app_locate(argv[1]);

		if (appschema == 0) {
			fprintf(stderr, "mpirun (app_locate): %s: ", argv[1]);
			lamfail("");
		}

		appd = app_parse(appschema, &lineno);
	}

	else {
/*
 * Parse the "command line" application specification.
 * Reconstruct its relevant parts from the parsed information.
 * It was parsed to distinguish between the appl/cmdline cases.
 *
 * Format: [<nodes>] program [-s <srcnode>] [-c #] [-- args]
 *
 */
		fmtbuf[0] = 0;

		if (index == 0) {
			if (nodeid == LOCAL) nodeid = getnodeid();
			noderanges(fmtbuf, nodeid, index);
		}

		for (--argc, ++argv; argc > 0; --argc, ++argv) {
			strcat(fmtbuf, *argv);
			strcat(fmtbuf, " ");
		}

		if (ao_taken(ad, "s")) {
			strcat(fmtbuf, "-s ");
			strcat(fmtbuf, ao_param(ad, "s", 0, 0));
			strcat(fmtbuf, " ");
		}
			
		if (fl_nprocs) {
			sprintf(smallbuf, "-c %d ", nprocs);
			strcat(fmtbuf, smallbuf);
		}

		ao_tail (ad, &argtailc, &argtailv);
		if (argtailc > 0) strcat(fmtbuf, "-- ");

		for ( ; argtailc > 0; --argtailc, ++argtailv) {
			strcat(fmtbuf, *argtailv);
			strcat(fmtbuf, " ");
		}
/*
 * Parse the command line.
 */
		appd = app_bufparse(fmtbuf, strlen(fmtbuf), &lineno);
		lineno = 0;
	}

	if (appd == 0) {
		if (lineno > 0) {
			fprintf(stderr, syntaxerr, lineno);
			errno = EUSAGE;
		} else {
			fprintf(stderr, "mpirun: cannot parse: ");
			terror("");
		}

		kexit(errno);
	}
/*
 * Pre-qualify the GER and set buffer limitations on each node.
 */
	if (!ao_taken(ad, "nger") && !ao_taken(ad, "c2c")) {
		if (app_ger(appd, TRUE)) kexit(errno);
	}
/*
 * Run the application.
 */
	if (app_run(appd, 0, rtf, fl_verbose, 1)) {
		kexit(errno);
	}
/*
 * Send process table to all processes.
 */
	if (sendproctbl(appd)) {
		terror("mpirun");
		app_doom(appd);
		kexit(errno);
	}
/*
 * If needed, wait for the application to terminate.
 */
	status = 0;
	
	if (fl_wait) {
		if (pwait(al_count(appd), &status)) {
			terror("mpirun (pwait)");
			app_doom(appd);
			kexit(errno);
		}
	}

	_kexit(status);
	return(0);
}

/*
 *	sendproctbl
 *
 *	Function:	- send process table to processes
 *	Accepts:	- app. schema desc.
 *	Returns:	- 0 or LAMERROR
 */
static int
sendproctbl(app)

LIST			*app;

{
	struct _gps	*procgps;		/* procs. GPS */
	int4		n;			/* favourite count */
	int4		err = 0;		/* error code */
/*
 * Allocate the process info array and fill it.
 */
	n = al_count(app);

	procgps = (struct _gps *) malloc((unsigned) n * sizeof(struct _gps));
	if (procgps == 0) return(LAMERROR);

	err = app_getprocs(app, procgps) || app_sendprocs(n, procgps);

	free((char *) procgps);
	return(err);
}

/*
 *	pwait
 *
 *	Function:	- waits for processes to complete
 *	Accepts:	- # of processes
 *			- return status of first child in error, else 0
 *	Returns:	- 0 or LAMERROR
 */
static int
pwait(nwait, childstat)

int4			nwait;
int			*childstat;

{
	int4		nodeid;			/* child's node ID */
	int		pid;			/* child's process ID */
	int		status;			/* return status */
	int		first = 1;		/* first error occurred */

	*childstat = 0;

	for ( ; nwait > 0; --nwait) {

		if (rpwait(&nodeid, &pid, &status)) return(LAMERROR);

		if (status) {
			printf("%d exited with status %d\n", pid, status);
			if (first) {
				*childstat = status;
				first = 0;
			}
		}
	}

	return(0);
}

/*
 *	setopmode
 *
 *	Function:	- set the operating mode
 */
static void
setopmode()

{
/*
 * flags
 */
	fl_fdesc = ao_taken(ad, "f");
	fl_verbose = ao_taken(ad, "v");
	fl_wait = ao_taken(ad, "w");
/*
 * followed options
 */
	nprocs = -1;
	fl_nprocs = ao_taken(ad, "c");
	if (fl_nprocs) ao_intparam(ad, "c", 0, 0, &nprocs);

	infile = ao_taken(ad, "i") ? ao_param(ad, "i", 0, 0) : (char *) 0;
	outfile = ao_taken(ad, "o") ? ao_param(ad, "o", 0, 0) : (char *) 0;
/*
 * runtime flags
 */
	rtf = RTF_MPIRUN;

	if (fl_wait) rtf |= RTF_WAIT;
	if (ao_taken(ad, "D")) rtf |= RTF_MPICWD;
	if (ao_taken(ad, "O")) rtf |= RTF_HOMOG;
	if (ao_taken(ad, "c2c")) rtf |= RTF_MPIC2C;
	if ( ! ao_taken(ad, "nger")) rtf |= RTF_MPIGER;

	if (ao_taken(ad, "t") || ao_taken(ad, "ton") || ao_taken(ad, "toff"))
		rtf |= RTF_TRACE;

	if ( ! ao_taken(ad, "toff")) rtf |= RTF_TRSWITCH;
}

/*
 *	setstdio
 *
 *	Function:	- set process stdio for remote filesystem
 *	Returns:	- 0 or LAMERROR
 */
static int
setstdio()

{
	int4		flags;			/* open file flags */
/*
 * Set stdin to input file or /dev/null.
 */
	if (infile) {
		sprintf(fmtbuf, "%d:%s", getnodeid(), infile);

		fd[0] = lam_rfopen(fmtbuf, LAM_O_RDONLY, 0);

		if (fd[0] < 0) return(LAMERROR);
	}

	_kio.ki_stdin = _ufd[fd[0]].fu_tfd;
/*
 * Set stdout to output file or user's tty.
 */
	if (outfile) {
		sprintf(fmtbuf, "%d:%s", getnodeid(), outfile);
		flags = LAM_O_WRONLY | LAM_O_CREAT | LAM_O_TRUNC;

		fd[1] = lam_rfopen(fmtbuf, flags, 0644);

		if (fd[1] < 0) return(LAMERROR);
	}
	else if (isatty(1) && !fl_fdesc) {
		sprintf(fmtbuf, "%d:%s", getnodeid(), ttyname(1));
		flags = LAM_O_WRONLY | LAM_O_REUSE;

		fd[1] = lam_rfopen(fmtbuf, flags, 0);

		if (fd[1] < 0) return(LAMERROR);
	}

	_kio.ki_stdout = _ufd[fd[1]].fu_tfd;
/*
 * Set stderr to user's tty.
 */
	if (isatty(2) && !fl_fdesc) {
		sprintf(fmtbuf, "%d:%s", getnodeid(), ttyname(2));
		flags = LAM_O_WRONLY | LAM_O_REUSE;

		fd[2] = lam_rfopen(fmtbuf, flags, 0);

		if (fd[2] < 0) return(LAMERROR);
	}

	_kio.ki_stderr = _ufd[fd[2]].fu_tfd;

	return(0);
}

/*
 *	noderanges
 *
 *	Function:	- fill buffer with list of nodes
 *			- compact nodes to ranges
 *			- formatting buffer assumed large enough
 *	Accepts:	- formatting buffer
 *			- initial node
 *			- initial index
 */
static void
noderanges(buf, initnode, initindex)

char			*buf;
int4			initnode;
int4			initindex;

{
	int4		node;			/* target node */
	int4		index;			/* for getnode() */
	int4		flags;			/* for getnode() */
	int4		next;			/* next expected node */
	int		fl_flush = 0;		/* flush node flag */

	sprintf(smallbuf, "n%d", initnode);
	strcat(buf, smallbuf);
/*
 * Loop over the nodes.
 */
	next = initnode + 1;

	nid_get(&index, &node, &flags);

	while (index != initindex) {
		if (node != next) {
			if (next > initnode + 1) {
/*
 * We have a "hole" ending a range of length >= 2.
 */
				sprintf(smallbuf, "-%d,%d", next - 1, node);
				strcat(buf, smallbuf);
			}
			else {
/*
 * We have a "hole" ending a range of length 1.
 */
				sprintf(smallbuf, ",%d", node);
				strcat(buf, smallbuf);
			}
			initnode = node;
			next = node + 1;
			fl_flush = 0;
		}
/*
 * Otherwise, keep incrementing.
 */
		else {
			++next;
			fl_flush = 1;
		}

		nid_get(&index, &node, &flags);
	}
/*
 * If we were still in a range, flush it out.
 */
	if (fl_flush) {
		sprintf(smallbuf, "-%d", next - 1);
		strcat(buf, smallbuf);
	}

	strcat(buf, " ");
}

/*
 *	help
 *
 *	Function:	- prints helpful information on this command
 */
static void
help()

{
	printf(
"\nSynopsis:\tmpirun [options] [<app>] [<prog> [<nodes>] [-- <args>]]\n");

	printf("\nDescription:\tStart an MPI application on LAM.\n");
	printf("\nOptions:\t-h\t\tPrint this help message.\n");
	printf("\t\t-c2c\t\tUse fast library (C2C) mode.\n");
	printf("\t\t-f\t\tDo not open stdio descriptors.\n");
	printf("\t\t-nger\t\tTurn off GER mode.\n");
	printf("\t\t-toff\t\tEnable tracing with generation off initially.\n");
	printf("\t\t-ton\t\tEnable tracing with generation on initially.\n");
	printf("\t\t-v\t\tBe verbose.\n");
	printf("\t\t-w\t\tWait for application to complete.\n");
	printf("\t\t-D\t\tSet current working directory to pathname.\n");
	printf("\t\t-O\t\tMulticomputer is homogeneous.\n");
	printf("\t\t-c <num>\tRun <num> copies of <prog>.\n");
	printf("\t\t-i <filename>\tRedirect standard input.\n");
	printf("\t\t-o <filename>\tRedirect standard output.\n");
	printf("\t\t-s <nodeid>\tLoad <prog> from this source node.\n");
	mnusage();
	printf("\t\th (local), o (origin), N (all)\n");
	printf("\nExample:        mpirun prog1 n0-7\n");
	printf("\t\t\t\"Execute `prog1' on nodes 0 thru 7.\"\n");
}
