/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	all_argv.c,v $
 * Revision 6.0  96/02/29  13:36:12  gdburns
 * Ohio Release
 * 
 *	Function:	- argument vector library
 */

#include <stdlib.h>
#include <string.h>

#include <portable.h>
#include <typical.h>

#define ARGSIZE		64

/*
 * global functions
 */
int			argvadd();
int			argvcount();
int			argvlen();
void			argvfree();
char			**argvbreak();
char			**argvdup();
char			*argvglue();

/*
 * backwards compatibility
 */
int addarg(argc, argv, arg) int *argc; char ***argv, *arg;
{ return(argvadd(argc, argv, arg)); }

/*
 *	argvadd
 *
 *	Function:	- extends an argv with a new arg
 *	Accepts:	- argc (value returned)
 *			- argv (value returned)
 *			- new arg
 */
int
argvadd(argc, argv, arg)

int			*argc;
char			***argv;
char			*arg;

{
/*
 * Create new argv.
 */
	if (*argv == 0) {
		*argv = (char **) malloc(2 * sizeof(char *));
		if (*argv == 0) return(LAMERROR);
		(*argv)[0] = 0;
		(*argv)[1] = 0;
	}
/*
 * Extend existing argv.
 */
	else {
		*argv = (char **) realloc((char *) *argv,
				(unsigned) (*argc + 2) * sizeof(char *));
		if (*argv == 0) return(LAMERROR);
	}

	(*argv)[*argc] = malloc((unsigned) strlen(arg) + 1);
	if ((*argv)[*argc] == 0) return(LAMERROR);

	strcpy((*argv)[*argc], arg);
	*argc = *argc + 1;
	(*argv)[*argc] = 0;
	return(0);
}

/*
 *	argvfree
 *
 *	Function:	- frees memory associated with argv
 *	Accepts:	- argv
 */
void
argvfree(argv)

char			**argv;

{
	char		**p;

	if (argv == 0) return;

	for (p = argv; *p; ++p) {
		free(*p);
	}

	free((char *) argv);
}

/*
 *	argvbreak
 * 
 *	Function:	- breaks single string into argv
 *			- allocates space for argv structure
 *			  which must be freed by the caller
 *	Accepts:	- argument string
 *			- arg delimiter within string
 *	Returns:	- argv or null ptr if error
 */
char **
argvbreak(args, delim)

CONST char		*args;
int			delim;

{
	char		arg[ARGSIZE];	/* single argument */
	char		**argv = 0;	/* argument structure ptr */
	CONST char	*p;		/* walk through args */
	char		*argtemp;	/* temporary single argument */
	int		argc = 0;	/* # arguments */
	int		arglen;		/* single argument length */

	while (*args) {
		p = args;
		arglen = 0;

		while ((*p != '\0') && (*p != delim)) {
			p++;
			arglen++;
		}
/*
 * zero length argument, skip
 */
		if (args == p) {
			p++;
		}
/*
 * tail argument, add straight from the original string
 */
		else if (*p == '\0') {

			if (argvadd(&argc, &argv, args)) return(0);
		}
/*
 * long argument, malloc buffer, copy and add
 */
		else if (arglen > (ARGSIZE - 1)) {
			argtemp = malloc((unsigned) (arglen + 1));
			if (argtemp == 0) return(0);

			strncpy(argtemp, args, arglen);
			argtemp[arglen] = '\0';

			if (argvadd(&argc, &argv, argtemp)) {
				free(argtemp);
				return(0);
			}

			free(argtemp);
		}
/*
 * short argument, copy to buffer and add
 */
		else {
			strncpy(arg, args, arglen);
			arg[arglen] = '\0';

			if (argvadd(&argc, &argv, arg)) return(0);
		}

		args = p;
	}

	return(argv);
}

/*
 *	argvcount
 * 
 *	Function:	- count the # of entries in argv
 *	Accepts:	- argv
 *	Returns:	- # of entries
 */
int
argvcount(argv)

char			**argv;

{
	char		**p;
	int		i;

	if (argv == 0) return(0);

	for (i = 0, p = argv; *p; i++, p++);

	return(i);
}

/*
 *	argvglue
 * 
 *	Function:	- glues argv into single string
 *			- delimiter and max string length specified
 *			- allocates space for string which must be
 *			  freed by the caller
 *	Accepts:	- argv
 *			- arg delimiter within string
 *			- max string length (or 0 meaning full argv)
 *	Returns:	- string or NULL
 */
char *
argvglue(argv, delim, maxlen)

char			**argv;
int			delim;
int			maxlen;

{
	char		**p;		/* favorite pointer */
	char		*pp;		/* another pointer */
	char		*str;		/* glued string */
	int		str_len = 0;	/* string length */
	int		i;		/* favorite index */
/*
 * Find the total string length in argv including delimiters.
 * The last delimiter is replaced by the NULL character.
 */
	for (p = argv; *p; p++) {
		str_len += strlen(*p) + 1;
	}

	if ((maxlen > 0) && (maxlen + 1 < str_len)) {
		str_len = maxlen + 1;
	}
/*
 * Allocate the string.
 */
	if ((str = malloc((unsigned) str_len)) == 0) return(str);
/*
 * Loop filling in the string.
 */
	str[--str_len] = '\0';
	p = argv;
	pp = *p;

	for (i = 0; i < str_len; i++) {

		if (*pp == '\0') {
/*
 * End of a string, fill in a delimiter and go to the next string.
 */
			str[i] = (char) delim;
			p++;
			pp = *p;
		} else {
			str[i] = *pp++;
		}
	}

	return(str);
}

/*
 *	argvlen
 * 
 *	Function:	- count the bytes consumed by an argv
 *	Accepts:	- argv
 *	Returns:	- length
 */
int
argvlen(argv)

char			**argv;

{
	char		**p;
	int		length;		/* argv length */

	if (argv == 0) return(0);

	length = sizeof(char *);

	for (p = argv; *p; p++) {
		length += strlen(*p) + 1 + sizeof(char *);
	}

	return(length);
}

/*
 *	argvdup
 *
 *	Function:	- create a duplicate copy of an argv
 *	Accepts:	- argv
 *	Returns:	- duplicate argv or NULL
 */
char **
argvdup(argv)

char			**argv;

{
	int		nargs;		/* count of arguments */
	int		len;		/* string length */
	int		i;		/* favourite index */
	char		*p;		/* favourite pointer */
	char		**dupv = 0;	/* duplicate argv */

	nargs = argvcount(argv);

	if (nargs > 0) {
		p = malloc((unsigned) argvlen(argv));

		if (p) {
			dupv = (char **) p;
			p += (nargs + 1) * sizeof(char *);

			for (i = 0; i < nargs; ++i) {
				len = strlen(argv[i]) + 1;
				memcpy(p, argv[i], len);
				dupv[i] = p;
				p += len;
			}

			dupv[nargs] = 0;
		}
	}

	return(dupv);
}
