/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	app_schema.c,v $
 * Revision 6.0  96/02/29  13:39:20  gdburns
 * Ohio Release
 * 
 *	Function:	- application schema routines
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <all_list.h>
#include <args.h>
#include <app_schema.h>
#include <bfreq.h>
#include <boot.h>
#include <freq.h>
#include <kio.h>
#include <ksignal.h>
#include <mpisys.h>
#include <net.h>
#include <portable.h>
#include <preq.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
LIST			*app_parse();
LIST			*app_bufparse();
void			app_doom();
void			app_free();
int			app_getprocs();
int			app_sendprocs();
int			app_ger();
int			app_run();
char			*app_locate();

/*
 * external variables
 */
extern int		argtailc;		/* from do_args() */
extern char		**argtailv;		/* from do_args() */
extern struct kio_t	_kio;			/* kernel I/O block */ 

/*
 * external functions
 */
extern char		*_path_find();

/*
 * local functions
 */
static int		parseline();
static void		doomprocs();
static void		fdincr();
static void		starterror();

/*
 *	app_parse
 *
 *	Function:	- parse an application schema
 *			- returns line # in case of parsing error
 *	Accepts:	- app. schema file 
 *			- ptr line # (returned value)
 *	Returns:	- app. schema descriptor or NULL
 */
LIST *
app_parse(appfile, plineno)

CONST char		*appfile;
int			*plineno;

{
	int		fd;			/* file desc. */
	int		nbytes;			/* # bytes read */
	int		bufsize;		/* buffer size */
	char		*filebuf;		/* file buffer */
	struct stat	fileinfo;		/* file information */
	LIST		*appd;			/* app. schema desc. */
/*
 * Get file info.
 */
	if (stat(appfile, &fileinfo)) return((LIST *) 0);

	bufsize = (int) fileinfo.st_size;
	if (bufsize <= 0) return((LIST *) 0);
/*
 * Allocate a file buffer.
 */
	filebuf = malloc((unsigned) bufsize);
	if (filebuf == 0) return((LIST *) 0);
/*
 * Open and read the file.
 */
	fd = open(appfile, O_RDONLY, 0);
	if (fd < 0) {
		free(filebuf);
		return((LIST *) 0);
	}

	nbytes = read(fd, filebuf, bufsize);
	close(fd);

	if (nbytes < bufsize) {
		if (nbytes >= 0) errno = EIO;
		free(filebuf);
		return((LIST *) 0);
	}
/*
 * Parse the app. schema buffer.
 */
	appd = app_bufparse(filebuf, bufsize, plineno);

	free(filebuf);
	return(appd);
}

/*
 *	app_bufparse
 *
 *	Function:	- parse an application schema from a buffer
 *			- returns line # in case of parsing error
 *	Accepts:	- buffer containing the app. schema
 *			- buffer size
 *			- ptr line # (returned value)
 *	Returns:	- app. schema descriptor or NULL
 */
LIST *
app_bufparse(appbuf, bufsize, pline)

char			*appbuf;
int			bufsize;
int			*pline;

{
	int		linesize;		/* size of parsed line */
	int		fl_comment;		/* comment mode flag */
	int		i;			/* favourite index */
	char		*p;			/* favourite pointer */
	LIST		*appd;			/* app. schema desc. */
/*
 * File must be pure ASCII.
 */
	*pline = 0;

	for (i = 0, p = appbuf; i < bufsize; ++i, ++p) {
		if ( ! isascii(*p)) {
			errno = ENOEXEC;
			return((LIST *) 0);
		}
	}
/*
 * Loop over the app. schema lines.
 */
	*pline = 1;
	fl_comment = 0;
	appd = 0;

	for ( ; bufsize > 0; ++appbuf, --bufsize) {
/*
 * Handle newline characters.
 */
		if (*appbuf == '\n') {
			++(*pline);
			fl_comment = 0;
		}
/*
 * Skip leading spaces and comments.
 */
		else if (isspace(*appbuf) || fl_comment) {
		}
/*
 * Skip comments.
 */
		else if (*appbuf == '#') {
			fl_comment = 1;
		}
/*
 * Parse the line.
 */
		else {
			linesize = parseline(&appd, appbuf, bufsize);

			if (linesize < 0) {
				app_free(appd);
				appd = 0;
				bufsize = 0;
			} else if (--linesize >= 0) {
				bufsize -= linesize;
				appbuf += linesize;
			}
		}
	}

	return(appd);
}

/*
 *	parseline
 *
 *	Function:	- parse a single app. schema line
 *			- update the app. schema desc.
 *	Accepts:	- ptr app. schema desc.
 *			- line buffer
 *			- buffer size
 *	Returns:	- parsed line size or LAMERROR
 */
static int
parseline(pappd, buf, size)

LIST			**pappd;
char			*buf;
int			size;

{
	int		lsize;			/* line size */
	int		cmdc;			/* # cmd line args */
	int		nprocs;			/* # processes */
	int4		srcnode;		/* source nodeid */
	int4		nodeid;			/* target node */
	int4		index;			/* for nid_get() */
	int4		nflags;			/* for nid_get() */
	char		*p;			/* favourite pointer */
	char		*nodev[3];		/* to create node lists */
	char		nodename[16];		/* node name */
	char		*cmdline;		/* command line buffer */
	char		**cmdv;			/* command line argv */
	struct apparg	*procargv;		/* process argv */
	struct appproc	proc;			/* process list entry */
/*
 * Copy the line to a temporary buffer.
 */
	p = strchr(buf, '\n');
	lsize = (p == 0) ? size : (int) (p - buf);

	if (lsize == 0) return(0);
/*
 * One extra char for terminator and 2 extra so we can put a fake argv[0]
 * at the start and then use the usual parsing routines.
 */
	cmdline = malloc((unsigned) lsize + 1 + 2);
	if (cmdline == 0) return(LAMERROR);
/*
 * Insert the fake argv[0] and then copy the line.
 */
	strcpy(cmdline, "x ");
	memcpy(cmdline + 2, buf, lsize);
	cmdline[lsize + 2] = '\0';
/*
 * Split it into a command line argv.
 */
	cmdv = argvbreak(cmdline, ' ');
	free(cmdline);

	if (cmdv == 0) return(LAMERROR);

	cmdc = argvcount(cmdv);
/*
 * Parse this command line.
 */
	reset_args();
	nid_free();

	followed("cs");
/*
 * Apart from the options, node specifications and the tail we should only
 * have the application name and the fake argument in the command line.
 */
	if (do_args(&cmdc, cmdv) || nid_parse(&cmdc, cmdv) || (cmdc != 2)) {
		free((char *) cmdv);
		errno = EUSAGE;
		return(LAMERROR);
	}
/*
 * Build the processes argv structure.
 */
	procargv = (struct apparg *) malloc(sizeof(struct apparg));
	if (procargv == 0) {
		free((char *) cmdv);
		return(LAMERROR);
	}

	procargv->apa_refcount = 0;
	procargv->apa_argc = argtailc + 1;

	if (argtailc > 0) {
		p = argtailv[-1];
		argtailv[-1] = cmdv[1];
		argtailv[argtailc] = 0;
		procargv->apa_argv = argvdup(argtailv - 1);
		argtailv[-1] = p;
	} else {
		p = cmdv[2];
		cmdv[2] = 0;
		procargv->apa_argv = argvdup(cmdv + 1);
		cmdv[2] = p;
	}

	free((char *) cmdv);

	if (procargv->apa_argv == 0) {
		free((char *) procargv);
		return(LAMERROR);
	}
/*
 * Loop adding the processes to the list.
 *
 * Handle the 4 cases:
 *	1) "foo -c # <nodes>"  # procs on this list of nodes
 *	2) "foo -c #"          # procs on each node in the LAM
 *	3) "foo <nodes>"       one proc on each of these nodes
 *	4) "foo"               one proc on the local node
 *
 * A negative "nprocs" is used in case 3.
 */

/*
 * Set the # of processes.
 */
	nid_get(&index, &nodeid, &nflags);

	if (opt_taken('c')) {
		intparam('c', &nprocs);
	} else {
		nprocs = (index < 0) ? 1 : -1;
	}
/*
 * Create a node list if needed.
 */
	if (index < 0) {
		nid_free();

		index = 2;
		nodev[0] = "nodes";
		nodev[2] = 0;

		if (opt_taken('c')) {
			nodev[1] = "N";
		} else {
			sprintf(nodename, "n%d", getnodeid());
			nodev[1] = nodename;
		}

		if (nid_parse(&index, nodev)) {
			free((char *) procargv->apa_argv);
			free((char *) procargv);
			return(LAMERROR);
		}

		nid_get(&index, &nodeid, &nflags);
	}
/*
 * Create an app. schema descriptor if needed.
 */
	if (*pappd == 0) {
		*pappd = al_init(sizeof(struct appproc), (int (*)()) 0);
		if (*pappd == 0) {
			free((char *) procargv->apa_argv);
			free((char *) procargv);
			return(LAMERROR);
		}
	}
/*
 * Parse the source nodeid.
 */
	if (opt_taken('s')) {
		srcnode = mnnodeid(getparam('s'));
		if (srcnode == NOTNODEID) return(LAMERROR);
	} else {
		srcnode = NOTNODEID;
	}
/*
 * Add the processes.
 */
	while (nprocs != 0) {
		proc.app_node = nodeid;
		proc.app_srcnode = (srcnode == NOTNODEID) ? nodeid : srcnode;
		proc.app_pid = -1;
		proc.app_args = procargv;

		if (al_append(*pappd, &proc) == 0) return(LAMERROR);

		++procargv->apa_refcount;

		nid_get(&index, &nodeid, &nflags);

		if (nprocs > 0) --nprocs;
		else if (index == 0) nprocs = 0;
	}

	return(lsize);
}

/*
 *	app_free
 *
 *	Function:	- free an application schema descriptor
 *	Accepts:	- app. schema desc.
 */
void
app_free(appd)

LIST			*appd;

{
	struct appproc	*pp;			/* ptr process entry */
	struct apparg	*pa;			/* ptr process args */

	if (appd == 0) return;

	for (pp = (struct appproc *) al_top(appd); pp;
			pp = (struct appproc *) al_next(appd, pp)) {
		pa = pp->app_args;

		if (--pa->apa_refcount == 0) {
			free((char *) pa);
		}
	}

	al_free(appd);
}

/*
 *	app_getprocs
 *
 *	Function:	- get process GPS (nodeid, pid, pindex)
 *			- given array must have >= nprocs entries
 *	Accepts:	- app. schema desc.
 *			- array of GPS structures (returned values)
 *	Returns:	- 0 or LAMERROR
 */
int
app_getprocs(appd, pgps)

LIST			*appd;
struct _gps		*pgps;

{
	struct appproc	*pp;			/* ptr process entry */

	if (appd == 0) {
		errno = EINVAL;
		return(LAMERROR);
	}

	for (pp = (struct appproc *) al_top(appd); pp;
			pp = (struct appproc *) al_next(appd, pp)) {
		pgps->gps_node = (pp->app_node == LOCAL) ?
				getnodeid() : pp->app_node;
		pgps->gps_pid = pp->app_pid;
		pgps->gps_idx = pp->app_idx;
		++pgps;
	}

	return(0);
}

/*
 *	app_sendprocs
 *
 *	Function:	- send GPS array to all processes in the array
 *	Accepts:	- number of processes in GPS array
 *			- GPS array
 *	Returns:	- 0 or LAMERROR
 */
int
app_sendprocs(nprocs, pgps)

int4			nprocs;
struct _gps		*pgps;

{
	struct nmsg	nhead;			/* network msg header */
	struct _gps	*p;			/* ptr GPS entry */
	int4		i;			/* favourite index */
/*
 * Set up the message.
 */
	nhead.nh_type = 0;
	nhead.nh_flags = DINT4MSG;
	nhead.nh_msg = (char *) pgps;
	nhead.nh_length = nprocs * sizeof(struct _gps);
/*
 * Loop sending to each process.
 */
	for (i = 0, p = pgps; i < nprocs; i++, p++) {
		nhead.nh_node = p->gps_node;
		nhead.nh_event = - p->gps_pid;

		if (nsend(&nhead)) return(LAMERROR);
	}
	
	return(0);
}

/*
 *	app_ger
 *
 *	Function:	- sets GER buffer limits for application
 *	Accepts:	- application schema
 *			- error printing flag
 */
int
app_ger(appd, flag_print)

LIST			*appd;
int			flag_print;

{
	int4		anode;			/* LOCAL filtered out */
	int4		nnodes;			/* # nodes in session */
	int4		*nodes;			/* the node IDs */
	int		i;
	int		*nprocs_node;		/* # procs on each node */
	int		nprocs;			/* # procs in application */
	struct appproc	*p;
	struct bfparms	bfp;			/* bufferd parameters */
/*
 * Get an array of nodeids and allocate a mirror array of process counts.
 */
	nnodes = getnall();
	nodes = (int4 *) malloc((unsigned) (nnodes * sizeof(int4)));
	if (getall(nodes, nnodes)) return(LAMERROR);
	nprocs_node = (int *) malloc((unsigned) (nnodes * sizeof(int)));
	if (nprocs_node == 0) return(LAMERROR);

	for (i = 0; i < nnodes; i++) nprocs_node[i] = 0;

	nprocs = al_count(appd);
/*
 * Tally the number of processes on each node.
 */
	p = (struct appproc *) al_top(appd);

	while (p) {
/*
 * Nodeids are not necessarily 0 to N-1, so we have to search for it.  (BFI)
 */
		for (i = 0; i < nnodes; ++i) {
			anode = (p->app_node == LOCAL) ? getnodeid() :
					p->app_node;

			if (anode == nodes[i]) break;
		}

		if (i >= nnodes) {
			errno = EBADNODE;

			if (flag_print) {
				fprintf(stderr, "mpirun: %s: ",
						mnemonic(p->app_node));
				terror("");
			}

			return(LAMERROR);
		}

		nprocs_node[i]++;
		p = (struct appproc *) al_next(appd, p);
	}
/*
 * Adjust the buffer limit on each node.
 */
	for (i = 0; i < nnodes; i++) {

		if (nprocs_node[i] == 0) continue;

		bfp.bfp_maxspace =
/*
 * the basic GER requirement
 */
			(MPI_GER * (MAXNMSGLEN + sizeof(struct nmsg)) *
			nprocs * nprocs_node[i]) +
/*
 * GER flow control messages - This one is tricky.  For processes
 * on your own node, you will only see the envelopes or the flow control
 * messages, but not max of both.  Since we have already covered local
 * processes above, we need not count them again here.
 */
			((MPI_GER - 1) * nprocs * (nprocs - nprocs_node[i]) *
			sizeof(struct nmsg)) +
/*
 * one packet for consumables to move through the buffer daemon
 */
			sizeof(struct nmsg) + MAXNMSGLEN +
/*
 * slush to cover up my mistakes
 */
			0x100000;

		bfp.bfp_maxspace = max(bfp.bfp_maxspace, BFDEFSPACE);

		if (rbfparms(nodes[i], &bfp)) {

			if (flag_print) {
				fprintf(stderr,
			    "mpirun: cannot allocate GER (%d bytes) on %s: ",
						bfp.bfp_maxspace,
						mnemonic(nodes[i]));
				terror("");
			}

			return(LAMERROR);
		}
	}

	free((char *) nodes);
	free((char *) nprocs_node);

	return(0);
}

/*
 *	app_run
 *
 *	Function:	- run the described application atomically
 *			- if a process cannot start, doom the running ones
 *	Accepts:	- app. schema desc.
 *			- number of parents in parent world
 *			- run-time environment flag
 *			- verbose flag
 *			- print error flag
 *	Returns:	- 0 or LAMERROR
 */
int
app_run(appd, nparent, rtf, verbose, prerror)

LIST			*appd;
int			nparent;
int4			rtf;
int			verbose;
int			prerror;

{
	struct appproc	*pp;			/* ptr process entry */
	int4		world_save;		/* saved ki_world */
	int4		parent_save;		/* saved ki_parent */
	int		i;			/* favourite index */
	int		pid;			/* process ID */
	int		idx;			/* process index */

	if ((appd == 0) || (al_count(appd) == 0)) {
		errno = EINVAL;
		return(LAMERROR);
	}
/*
 * Set up the kernel I/O block.
 */
	world_save = _kio.ki_world;
	parent_save = _kio.ki_parent;

	_kio.ki_parent = nparent;
	_kio.ki_world = al_count(appd);
/*
 * Increment the stdio counts.
 */
	fdincr(_kio.ki_world);
/*
 * Loop starting each process.
 */
	pp = (struct appproc *) al_top(appd);

	for (i = 0; pp; ++i, pp = (struct appproc *) al_next(appd, pp)) {
/*
 * Start the process and store its pid.
 */
		if (rploadgo(pp->app_srcnode, pp->app_node,
				pp->app_args->apa_argv[0], rtf,
				pp->app_args->apa_argv, &pid, &idx)) {
			starterror(pp->app_args->apa_argv[0], pp->app_node,
					prerror);
			break;
		}

		pp->app_pid = pid;
		pp->app_idx = idx;
		
		if (verbose) {
			printf("%d %s running on %s\n", pid,
					pp->app_args->apa_argv[0],
					mnemonic(pp->app_node));
		}
	}

	_kio.ki_world = world_save;
	_kio.ki_parent = parent_save;
/*
 * If all processes were not started, cleanup and doom the running ones.
 */
	world_save = al_count(appd);

	if (i < world_save) {
		fdincr(- world_save);
		doomprocs(appd, i);
		return(LAMERROR);
	}

	return(0);
}

/*
 *	app_doom
 *
 *	Function:	- doom all application processes
 *	Accepts:	- app. schema desc.
 */
void
app_doom(appd)

LIST			*appd;

{
	if (appd) doomprocs(appd, al_count(appd));
}

/*
 *	app_locate
 *
 *	Function:	- locates the application schema
 *	Accepts:	- filename
 *	Returns:	- full pathname or NULL
 */
char *
app_locate(filename)

char			*filename;

{
	int		pathc = 0;		/* paths argc */
	char		**pathv = 0;		/* paths argv */
	char		*appdir;		/* application dir */
	char		*fullpath;		/* full pathname */

	appdir = getenv("LAMAPPLDIR");
	if (appdir) argvadd(&pathc, &pathv, appdir);

	argvadd(&pathc, &pathv, "");
	argvadd(&pathc, &pathv, "$LAMHOME/boot");
	argvadd(&pathc, &pathv, "$TROLLIUSHOME/boot");
	argvadd(&pathc, &pathv, DEFPSCHEMA);

	fullpath = _path_find(filename, pathv, R_OK);
	argvfree(pathv);
	return(fullpath);
}

/*
 *	doomprocs
 *
 *	Function:	- doom the first N processes in an application
 *	Accepts:	- app. schema desc.
 *			- # procs to doom
 */
static void
doomprocs(appdesc, ndoom)

LIST			*appdesc;
int4			ndoom;

{
	struct appproc	*p;

	p = (struct appproc *) al_top(appdesc);

	for ( ; ndoom > 0; --ndoom) {
		rpdoom(p->app_node, SELECT_PID, p->app_pid, SIGUDIE);
		p = (struct appproc *) al_next(appdesc, p);
	}
}

/*
 *	fdincr
 *
 *	Function:	- increment stdio file desc.
 *	Accepts:	- increment value
 */
static void
fdincr(incr)

int4			incr;

{
	if (_kio.ki_stdin > 2) lam_rfincr(0, incr);
	if (_kio.ki_stdout > 2) lam_rfincr(1, incr);
	if (_kio.ki_stderr > 2) lam_rfincr(2, incr);
}

/*
 *	starterror
 *
 *	Function:	- print can't start program error
 *	Accepts:	- file name
 *			- ID of node failed on
 *			- print error flag
 */
static void
starterror(file, nodeid, prerror)

char			*file;
int			nodeid;
int			prerror;

{
	if (prerror) {
		fprintf(stderr, "mpirun: cannot start %s on %s: ",
				file, mnemonic(nodeid));
		terror("");
	}
}
