/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	alltoall.c,v $
 * Revision 6.0  96/03/03  23:03:07  nevin
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/03/03  23:01:48  nevin
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:09  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:03:45  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:12  gdburns
 * Ohio Release
 * 
 *	Function:	- send/recv data from each node to all nodes
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- recv buffer
 *			- recv count
 *			- recv datatype
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * local functions
 */
static int		c2c_alltoall();
static int		lamd_alltoall();


int
MPI_Alltoall(sbuf, scount, sdtype, rbuf, rcount, rdtype, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
MPI_Comm		comm;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPIALLTOALL);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIALLTOALL,
			__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if ((sdtype == MPI_DATATYPE_NULL) || (rdtype == MPI_DATATYPE_NULL)) {
		return(__lam_errfunc_(comm, BLKMPIALLTOALL,
			__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if ((scount < 0) || (rcount < 0)) {
		return(__lam_errfunc_(comm, BLKMPIALLTOALL,
			__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	_LAM_TRACE(__lam_trcffstart_(BLKMPIALLTOALL));

	return(RPI_SPLIT(lamd_alltoall, c2c_alltoall,
		(sbuf, scount, sdtype, rbuf, rcount, rdtype, comm)));
}

/*
 *	c2c_alltoall
 *
 *	Function:	- MPI_Alltoall for the C2C RPI
 *	Accepts:	- same as MPI_Alltoall
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */
static int
c2c_alltoall(sbuf, scount, sdtype, rbuf, rcount, rdtype, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		rank;			/* my rank */
	int		size;			/* group size */
	int		nreqs;			/* # requests */
	int		err;			/* error code */
	char		*psnd;			/* ptr send buffer */
	char		*prcv;			/* ptr recv buffer */
	MPI_Aint	sndinc;			/* send increment */
	MPI_Aint	rcvinc;			/* rcv increment */
	MPI_Request	*req;			/* request array */
	MPI_Request	*preq;			/* ptr request */
	MPI_Status	*stat;			/* status array */
/*
 * Initialize.
 */
	MPI_Comm_size(comm, &size);
	MPI_Comm_rank(comm, &rank);

	MPI_Type_extent(sdtype, &sndinc);
	MPI_Type_extent(rdtype, &rcvinc);

	sndinc *= scount;
	rcvinc *= rcount;
/*
 * Allocate arrays of requests and status structures.
 */
	nreqs = 2 * (size - 1);

	if (nreqs > 0) {
		req = (MPI_Request *) malloc((unsigned)
						nreqs * sizeof(MPI_Request));
		stat = (MPI_Status *) malloc((unsigned)
						nreqs * sizeof(MPI_Status));

		if ((req == 0) || (stat == 0)) {
			if (req) free((char *) req);
			if (stat) free((char *) stat);
			return(__lam_errfunc_(comm, BLKMPIALLTOALL,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
	}
	else {
		req = 0;
		stat = 0;
	}
/*
 * Switch to collective communicator.
 */
	__lam_mkcoll_(comm);
/*
 * simple optimization
 */
	psnd = ((char *) sbuf) + (rank * sndinc);
	prcv = ((char *) rbuf) + (rank * rcvinc);

	err = __lam_dtsndrcv_(psnd, scount, sdtype,
			prcv, rcount, rdtype, BLKMPIALLTOALL, comm);

	if (err != MPI_SUCCESS) {
		if (req) free((char *) req);
		if (stat) free((char *) stat);
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
	}
/*
 * If only one process, generate run time trace and we're done.
 */
	if (size == 1) {
		__lam_mkpt_(comm);
		_LAM_TRACE(__lam_trcffend_(BLKMPIALLTOALL, -1, comm,
				sdtype, scount));
		
		__lam_resetfunc_(BLKMPIALLTOALL);
		return(MPI_SUCCESS);
	}
/*
 * Initiate all send/recv to/from others.
 */
	psnd = (char *) sbuf;
	prcv = (char *) rbuf;
	preq = req;

	for (i = 0; i < size; ++i, prcv += rcvinc) {

		if (i == rank) continue;

		err = MPI_Recv_init(prcv, rcount, rdtype, i,
					BLKMPIALLTOALL, comm, preq++);

		if (err != MPI_SUCCESS) {
			free((char *) req);
			free((char *) stat);
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
		}
	}

	for (i = 0; i < size; ++i, psnd += sndinc) {

		if (i == rank) continue;

		err = MPI_Send_init(psnd, scount, sdtype, i,
					BLKMPIALLTOALL, comm, preq++);

		if (err != MPI_SUCCESS) {
			free((char *) req);
			free((char *) stat);
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
		}
	}
/*
 * Start all the requests.
 */
	err = MPI_Startall(nreqs, req);

	if (err != MPI_SUCCESS) {
		free((char *) req);
		free((char *) stat);
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
	}
/*
 * Wait for them all.
 */
	err = MPI_Waitall(nreqs, req, stat);

	__lam_mkpt_(comm);
	free((char *) stat);

	if (err != MPI_SUCCESS) {
		free((char *) req);
		return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
	}
/*
 * Free the requests.
 */
	for (i = 0, preq = req; i < nreqs; ++i, ++preq) {

		err = MPI_Request_free(preq);

		if (err != MPI_SUCCESS) {
			free((char *) req);
			return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
		}
	}

	free((char *) req);

	_LAM_TRACE(__lam_trcffend_(BLKMPIALLTOALL, -1, comm, sdtype, scount));

	__lam_resetfunc_(BLKMPIALLTOALL);
	return(MPI_SUCCESS);
}


/*
 *	lamd_alltoall
 *
 *	Function:	- MPI_Alltoall for the LAMD RPI
 *	Accepts:	- same as MPI_Alltoall
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */
static int
lamd_alltoall(sbuf, scount, sdtype, rbuf, rcount, rdtype, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			rank;			/* my rank */
    int			size;			/* group size */
    int			err;			/* error code */
    char		*psnd;			/* ptr send buffer */
    char		*prcv;			/* ptr recv buffer */
    MPI_Aint		sndinc;			/* send increment */
    MPI_Aint		rcvinc;			/* rcv increment */
    MPI_Status		stat;			/* status */
/*
 * Initialize.
 */
    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);

    MPI_Type_extent(sdtype, &sndinc);
    MPI_Type_extent(rdtype, &rcvinc);

    sndinc *= scount;
    rcvinc *= rcount;
/*
 * Switch to collective communicator.
 */
    __lam_mkcoll_(comm);
/*
 * simple optimization
 */
    psnd = ((char *) sbuf) + (rank * sndinc);
    prcv = ((char *) rbuf) + (rank * rcvinc);

    err = __lam_dtsndrcv_(psnd, scount, sdtype,	prcv, rcount, rdtype,
			BLKMPIALLTOALL, comm);

    if (err != MPI_SUCCESS) {
	__lam_mkpt_(comm);
	return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
    }
/*
 * Do sendrecv's with others if any.
 */
    psnd = (char *) sbuf;
    prcv = (char *) rbuf;

    for (i = 0; i < size; ++i, prcv += rcvinc, psnd += sndinc) {

	if (i == rank) continue;
	
	err = MPI_Sendrecv(psnd, scount, sdtype, i, BLKMPIALLTOALL,
		prcv, rcount, rdtype, i, BLKMPIALLTOALL, comm, &stat);

	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIALLTOALL, err));
	}
    }

    __lam_mkpt_(comm);
    _LAM_TRACE(__lam_trcffend_(BLKMPIALLTOALL, -1, comm, sdtype, scount));

    __lam_resetfunc_(BLKMPIALLTOALL);
    return(MPI_SUCCESS);
}
