/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	barrier.c,v $
 * Revision 6.0  96/02/29  14:11:59  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:06  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:49  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:16  gdburns
 * Ohio Release
 * 
 *	Function:	- barrier synchronization
 *	Accepts:	- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>

/*
 * local functions
 */
static int		barrier_lin();
static int		barrier_log();

int
MPI_Barrier(comm)

MPI_Comm		comm;

{
	int		size;			/* group size */
/*
 * Prepare error handler if needed.
 */
	__lam_initerr_();
	__lam_setfunc_(BLKMPIBARRIER);
/*
 * Check for invalid arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(comm,
			BLKMPIBARRIER, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPIBARRIER, __lam_mkerr_(MPI_ERR_COMM, 0)));
        }
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIBCAST);
/*
 * Decide which algorithm to use.
 */
	MPI_Comm_size(comm, &size);

	if (size <= LAM_COLLMAXLIN) {
		return(barrier_lin(comm));
	} else {
		return(barrier_log(comm));
	}
}

/*
 *	barrier_lin
 *
 *	Function:	- barrier using O(N) algorithm
 *	Accepts:	- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
barrier_lin(comm)

MPI_Comm		comm;

{
	int		size;			/* group size */
	int		rank;			/* caller rank */
	int		err;			/* error code */
	int		i;			/* favourite index */
	MPI_Status	stat;			/* receive status */

	MPI_Comm_size (comm, &size);
	MPI_Comm_rank(comm, &rank);

	__lam_mkcoll_(comm);
/*
 * All non-root send & receive zero-length message.
 */
	if (rank > 0) {
		err = MPI_Send((void *) 0, 0, MPI_BYTE,
					0, BLKMPIBARRIER, comm);

		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIBARRIER, err));
		}

		err = MPI_Recv((void *) 0, 0, MPI_BYTE,
					0, BLKMPIBARRIER, comm, &stat);

		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIBARRIER, err));
		}
	}
/*
 * The root collects and broadcasts the messages.
 */
	else {
		for (i = 1; i < size; ++i) {

			err = MPI_Recv((void *) 0, 0, MPI_BYTE, MPI_ANY_SOURCE,
						BLKMPIBARRIER, comm, &stat);

			if (err != MPI_SUCCESS) {
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIBARRIER, err));
			}
		}

		for (i = 1; i < size; ++i) {

			err = MPI_Send((void *) 0, 0, MPI_BYTE,
						i, BLKMPIBARRIER, comm);

			if (err != MPI_SUCCESS) {
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIBARRIER, err));
			}
		}
	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIBCAST, -1, comm, 0, 0);
	
	__lam_resetfunc_(BLKMPIBARRIER);
	return(MPI_SUCCESS);
}

/*
 *	barrier_log
 *
 *	Function:	- barrier using O(log(N)) algorithm
 *	Accepts:	- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
barrier_log(comm)

MPI_Comm		comm;

{
	int		size;			/* group size */
	int		rank;			/* caller rank */
	int		peer;			/* peer rank */
	int		dim;			/* cube dimension */
	int		hibit;			/* high ON bit position */
	int		mask;			/* rank bit mask */
	int		err;			/* error code */
	int		i;			/* favourite index */
	MPI_Status	stat;			/* receive status */

	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);
 
	__lam_mkcoll_(comm);

	dim = comm->c_cube_dim;
	hibit = __lam_hibit_(rank, dim);
	--dim;
/*
 * Send null-messages up and down the tree.
 * Synchronization at the root (rank 0).
 */
/*
 * Receive from children.
 */
	for (i = dim, mask = 1 << i; i > hibit; --i, mask >>= 1) {

		peer = rank | mask;
		if (peer < size) {
			err = MPI_Recv((void *) 0, 0, MPI_BYTE,
					peer, BLKMPIBARRIER, comm, &stat);

			if (err != MPI_SUCCESS) {
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIBARRIER, err));
			}
		}
	}
/*
 * Send to and receive from parent.
 */
	if (rank > 0) {

		peer = rank & ~(1 << hibit);

		err = MPI_Send((void *) 0, 0, MPI_BYTE,
						peer, BLKMPIBARRIER, comm);
		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIBARRIER, err));
		}

		err = MPI_Recv((void *) 0, 0, MPI_BYTE, peer,
						BLKMPIBARRIER, comm, &stat);
		if (err != MPI_SUCCESS) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIBARRIER, err));
		}
	}
/*
 * Send to children.
 */
	for (i = hibit + 1, mask = 1 << i; i <= dim; ++i, mask <<= 1) {

		peer = rank | mask;
		if (peer < size) {
			err = MPI_Send((void *) 0, 0, MPI_BYTE,
						peer, BLKMPIBARRIER, comm);

			if (err != MPI_SUCCESS) {
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIBARRIER, err));
			}
		}
	}

	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIBCAST, -1, comm, 0, 0);

	__lam_resetfunc_(BLKMPIBARRIER);
	return(MPI_SUCCESS);
}
