/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	bcast.c,v $
 * Revision 6.0.1.1  96/04/08  18:07:45  nevin
 * Added missing trace generation statements
 * 
 * Revision 6.0  96/02/29  14:12:00  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:18:14  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:11  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:03:49  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:17  gdburns
 * Ohio Release
 * 
 *	Function:	- broadcast info to all nodes SIMD-style
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * local functions
 */
static int		bcast_lin();
static int		bcast_log();
static int		bcast_lin_lamd();
static int		bcast_log_lamd();

int
MPI_Bcast(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			size;			/* group size */
    struct _gps		*p;			/* favourite pointer */
/*
 * Prepare error handler if needed.
 */
    __lam_initerr_();
    __lam_setfunc_(BLKMPIBCAST);
/*
 * Check for invalid arguments.
 */
    if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
	return(__lam_errfunc_(comm, BLKMPIBCAST,
	    		__lam_mkerr_(MPI_ERR_COMM, 0)));
    }

    if (datatype == MPI_DATATYPE_NULL) {
	return(__lam_errfunc_(comm, BLKMPIBCAST,
	    		__lam_mkerr_(MPI_ERR_TYPE, 0)));
    }

    if (count < 0) {
	return(__lam_errfunc_(comm, BLKMPIBCAST,
	    		__lam_mkerr_(MPI_ERR_COUNT, 0)));
    }

    MPI_Comm_size(comm, &size);

    if ((root >= size) || (root < 0)) {
	return(__lam_errfunc_(comm, BLKMPIBCAST,
	    		__lam_mkerr_(MPI_ERR_ROOT, 0)));
    }

    _LAM_TRACE(__lam_trcffstart_(BLKMPIBCAST));
/*
 * Set debugging parameters.
 */
    p = &(comm->c_group->g_procs[root]->p_gps);

    __lam_setparam_(BLKMPIBCAST, root | (p->gps_grank << 16),
	(p->gps_node << 16) | p->gps_idx);
/*
 * Decide which algorithm to use.
 */
    if (size <= 1) {
	__lam_resetfunc_(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
    else if (size <= LAM_COLLMAXLIN) {
	return(RPI_SPLIT(bcast_lin_lamd, bcast_lin,
	    		(buff, count, datatype, root, comm)));
    }
    else {
	return(RPI_SPLIT(bcast_log_lamd, bcast_log,
	    		(buff, count, datatype, root, comm)));
    }
}

/*
 *	bcast_lin
 *
 *	Function:	- broadcast using O(N) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_lin(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			err;			/* error code */
    MPI_Request		*preq;			/* ptr request */
    MPI_Request		reqs[LAM_COLLMAXLIN];	/* requests */
    MPI_Status		stats[LAM_COLLMAXLIN];	/* status structures */

    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);

    __lam_mkcoll_(comm);
/*
 * Non-root receive the data.
 */
    if (rank != root) {
	err = MPI_Recv(buff, count, datatype, root, BLKMPIBCAST, comm, stats);
	__lam_mkpt_(comm);

	if (err != MPI_SUCCESS) {
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}

	_LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
	__lam_resetfunc_(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
/*
 * Root sends data to all others.
 */
    for (i = 0, preq = reqs; i < size; ++i) {

	if (i == rank) continue;

	err = MPI_Send_init(buff, count, datatype, i, BLKMPIBCAST,
	    			comm, preq++);

	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Start and wait on all requests.
 */
    err = MPI_Startall(size - 1, reqs);
    if (err != MPI_SUCCESS) {
	__lam_mkpt_(comm);
	return(__lam_errfunc_(comm, BLKMPIBCAST, err));
    }

    err = MPI_Waitall(size - 1, reqs, stats);
    __lam_mkpt_(comm);

    if (err != MPI_SUCCESS) {
	return(__lam_errfunc_(comm, BLKMPIBCAST, err));
    }
/*
 * Free the requests.
 */
    for (i = 0, preq = reqs; i < size; ++i) {

	if (i == rank) continue;

	err = MPI_Request_free(preq);
	if (err != MPI_SUCCESS) {
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}

	++preq;
    }

    _LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
    __lam_resetfunc_(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_log
 *
 *	Function:	- broadcast using O(log(N)) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_log(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			vrank;			/* caller virtual rank */
    int			peer;			/* peer rank */
    int			dim;			/* cube dimension */
    int			hibit;			/* high ON bit position */
    int			mask;			/* rank bit mask */
    int			err;			/* error code */
    int			nreqs;			/* # requests */
    MPI_Request		*preq;			/* ptr request */
    MPI_Request		reqs[LAM_COLLMAXDIM];	/* requests */
    MPI_Status		stats[LAM_COLLMAXDIM];	/* status structures */

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    __lam_mkcoll_(comm);

    vrank = (rank + size - root) % size;

    dim = comm->c_cube_dim;
    hibit = __lam_hibit_(vrank, dim);
    --dim;
/*
 * Receive data from parent in the tree.
 */
    if (vrank > 0) {
	peer = ((vrank & ~(1 << hibit)) + root) % size;

	err = MPI_Recv(buff, count, datatype, peer, BLKMPIBCAST, comm, stats);
	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Send data to the children.
 */
    preq = reqs;
    nreqs = 0;

    for (i = hibit + 1, mask = 1 << i; i <= dim; ++i, mask <<= 1) {

	peer = vrank | mask;
	if (peer < size) {
	    peer = (peer + root) % size;
	    ++nreqs;

	    err = MPI_Send_init(buff, count, datatype, peer, BLKMPIBCAST,
				comm, preq++);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }
/*
 * Start and wait on all requests.
 */
    if (nreqs > 0) {
	err = MPI_Startall(nreqs, reqs);
	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}

	err = MPI_Waitall(nreqs, reqs, stats);
	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}

	for (i = 0, preq = reqs; i < nreqs; ++i, ++preq) {

	    err = MPI_Request_free(preq);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }

    __lam_mkpt_(comm);
    _LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
    __lam_resetfunc_(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_lin_lamd
 *
 *	Function:	- LAMD broadcast using O(N) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_lin_lamd(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			longproto;		/* long protocol */
    int			err;			/* error code */
    MPI_Status		stat;			/* status */

    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);

    longproto = (datatype->dt_size * count) > MAXNMSGLEN;
    __lam_mkcoll_(comm);
/*
 * Non-root receives the data.
 */
    if (rank != root) {
	if (longproto) {
	    err = MPI_Recv(buff, 0, MPI_BYTE, root, BLKMPIBCAST, comm, &stat);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }

	    err = MPI_Send(buff, 0, MPI_BYTE, root, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}

	err = MPI_Recv(buff, count, datatype, root, BLKMPIBCAST, comm, &stat);

	__lam_mkpt_(comm);
	if (err != MPI_SUCCESS) {
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}

	_LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
	__lam_resetfunc_(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
/*
 * Root sends data to all others.
 */
    if (longproto) {
	for (i = 0; i < size; ++i) {

	    if (i == rank) continue;

	    err = MPI_Send(buff, 0, MPI_BYTE, i, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
/*
 * Wait on ack messages and send data.
 */
	for (i = 0; i < size - 1; i++) {

	    err = MPI_Recv(buff, 0, MPI_BYTE, MPI_ANY_SOURCE, BLKMPIBCAST,
				comm, &stat);

	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
		
	    err = MPI_Send(buff, count, datatype, stat.MPI_SOURCE,
				BLKMPIBCAST, comm);
		
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }
    else {
	for (i = 0; i < size; i++) {
	    if (i == root) continue;

	    err = MPI_Send(buff, count, datatype, i, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }
    
    __lam_mkpt_(comm);
    _LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
    __lam_resetfunc_(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_log_lamd
 *
 *	Function:	- LAMD broadcast using O(log(N)) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_log_lamd(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			vrank;			/* caller virtual rank */
    int			peer;			/* peer rank */
    int			dim;			/* cube dimension */
    int			hibit;			/* high ON bit position */
    int			mask;			/* rank bit mask */
    int			longproto;		/* long protocol */
    int			err;			/* error code */
    int			nreqs;			/* # requests */
    MPI_Status		stat;			/* status structures */

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    __lam_mkcoll_(comm);
    longproto = (datatype->dt_size * count) > MAXNMSGLEN;

    vrank = (rank + size - root) % size;

    dim = comm->c_cube_dim;
    hibit = __lam_hibit_(vrank, dim);
    --dim;
/*
 * Receive data from parent in the tree.
 */
    if (vrank > 0) {
	peer = ((vrank & ~(1 << hibit)) + root) % size;

	if (longproto) {
	    err = MPI_Recv(buff, 0, MPI_BYTE, peer, BLKMPIBCAST, comm, &stat);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }

	    err = MPI_Send(buff, 0, MPI_BYTE, peer, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}

	err = MPI_Recv(buff, count, datatype, peer, BLKMPIBCAST, comm, &stat);
	if (err != MPI_SUCCESS) {
	    __lam_mkpt_(comm);
	    return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Send data or primer message to the children.
 */
    nreqs = 0;

    for (i = hibit + 1, mask = 1 << i; i <= dim; ++i, mask <<= 1) {

	peer = vrank | mask;
	if (peer < size) {
	    peer = (peer + root) % size;
	    ++nreqs;

	    if (longproto) {
		err = MPI_Send(buff, 0, MPI_BYTE, peer, BLKMPIBCAST, comm);
	    } else {
		err = MPI_Send(buff, count, datatype, peer, BLKMPIBCAST, comm);
	    }
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }

    if (longproto) {
/*
 * Wait for primer ack messages and send data.
 */
	for (i = 0; i < nreqs; i++) {

	    err = MPI_Recv(buff, 0, MPI_BYTE, MPI_ANY_SOURCE, BLKMPIBCAST,
				comm, &stat);

	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
		
	    err = MPI_Send(buff, count, datatype, stat.MPI_SOURCE,
				BLKMPIBCAST, comm);
		
	    if (err != MPI_SUCCESS) {
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIBCAST, err));
	    }
	}
    }

    __lam_mkpt_(comm);
    _LAM_TRACE(__lam_trcffend_(BLKMPIBCAST, root, comm, datatype, count));
    __lam_resetfunc_(BLKMPIBCAST);
    return(MPI_SUCCESS);
}
