/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	cartcreate.c,v $
 * Revision 6.0  96/02/29  14:12:06  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:18:53  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:03:54  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:22  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:03  gdburns
 * new for 5.1
 * 
 *	Function:	- create a communicator with cartesian mapping
 *	Accepts:	- old communicator
 *			- # dimensions
 *			- # procs per dimension
 *			- dimension periodicity
 *			- rank reordering flag
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Cart_create(comm, ndims, dims, periods, reorder, pnewcomm)

MPI_Comm		comm;
int			ndims;
int			*dims;
int			*periods;
int			reorder;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;
	MPI_Group	newgroup;
	int		size;
	int		nprocs;
	int		rank;
	int		err;
	int		range[1][3];
	int		i;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPICARTCREATE);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPICARTCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCREATE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (ndims < 1) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCREATE, __lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if ((dims == 0) || (periods == 0) || (pnewcomm == 0)) {
		return(__lam_errfunc_(comm,
			BLKMPICARTCREATE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPICARTCREATE);
/*
 * Compute the # of processes in the grid.
 */
	nprocs = 1;
	for (i = 0, p = dims; i < ndims; ++i, ++p) {
		if (*p <= 0) {
			return(__lam_errfunc_(comm, BLKMPICARTCREATE,
					__lam_mkerr_(MPI_ERR_DIMS, 0)));
		}

		nprocs *= *p;
	}
/*
 * Create the group for the new communicator.
 */
	err = MPI_Comm_size(comm, &size);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
	}

	if (nprocs > size) {
		return(__lam_errfunc_(comm, BLKMPICARTCREATE,
				__lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if (nprocs == size) {
		err = MPI_Comm_group(comm, &newgroup);
	} else {
		range[0][0] = 0;
		range[0][1] = nprocs - 1;
		range[0][2] = 1;
		err = MPI_Group_range_incl(comm->c_group, 1, range, &newgroup);
	}

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
	}
/*
 * Create the new communicator.
 */
	err = MPI_Comm_create(comm, newgroup, pnewcomm);
	if (err != MPI_SUCCESS) {
		MPI_Group_free(&newgroup);
		return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
	}
/*
 * Fill the communicator with topology information. 
 */
	newcomm = *pnewcomm;
	if (newcomm != MPI_COMM_NULL) {
		newcomm->c_topo_type = MPI_CART;
		newcomm->c_topo_nprocs = nprocs;
		newcomm->c_topo_ndims = ndims;

		newcomm->c_topo_dims = (int *)
				malloc((unsigned) 2 * ndims * sizeof(int));
		if (newcomm->c_topo_dims == 0) {
			return(__lam_errfunc_(comm, BLKMPICARTCREATE,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
		newcomm->c_topo_coords = newcomm->c_topo_dims + ndims;

		for (i = 0, p = newcomm->c_topo_dims; i < ndims; ++i, ++p) {
			*p = (*periods) ? -(*dims) : *dims;
			++dims;
			++periods;
		}
/*
 * Compute the caller's coordinates.
 */
		err = MPI_Comm_rank(newcomm, &rank);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
		}

		err = MPI_Cart_coords(newcomm, rank,
					ndims, newcomm->c_topo_coords);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
		}
	}

	err = MPI_Group_free(&newgroup);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICARTCREATE, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPICARTCREATE, -1, comm, 0, 0);

	__lam_resetfunc_(BLKMPICARTCREATE);
	return(MPI_SUCCESS);
}
