/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	csplit.c,v $
 * Revision 6.0  96/02/29  14:12:19  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:20:45  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:05  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:37  gdburns
 * Ohio Release
 * 
 *	Function:	- split communicator into disjoint subsets
 *	Accepts:	- old communicator
 *			- process colour
 *			- process key
 *			- ptr new communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern int		__lam_cubedim_();
extern int		__lam_trcomm_();

/*
 * local functions
 */
static int		cmpprocs();

int
MPI_Comm_split(comm, colour, key, pnewcomm)

MPI_Comm		comm;
int			colour;
int			key;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;		/* new communicator */
	MPI_Group	oldgroup;		/* old group */
	MPI_Group	newgroup;		/* new group */
	int		mycid;			/* my context ID */
	int		cid;			/* global context ID */
	int		i;			/* favourite index */
	int		err;			/* error code */
	int		myrank;			/* my group rank */
	int		nprocs;			/* # processes */
	int		mynprocs;		/* # procs in new comm */
	int		*procs;			/* procs array */
	int		*myprocs;		/* new comm procs array */
	int		*p, *p2;		/* favourite pointers */

	__lam_initerr_();
	__lam_setfunc_(BLKMPICOMMSPLIT);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (pnewcomm == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	_LAM_TRACE(__lam_trcffstart_(BLKMPICOMMSPLIT));
/*
 * Create the new context ID using MPI_Allreduce().
 * Processes not in group participate but do not affect the context ID.
 */
	MPI_Comm_rank(comm, &myrank);

	mycid = (myrank == MPI_UNDEFINED) ? 0 : __lam_getcid_();
	if (mycid < 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}

	err = MPI_Allreduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, comm);
	if (err != MPI_SUCCESS) {
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMSPLIT, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMSPLIT);
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT, err));
	}
/*
 * Create the array of process information.
 */
	MPI_Comm_size(comm, &nprocs);

	procs = (int *) malloc((unsigned) 3 * nprocs * sizeof(int));
	if (procs == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Gather all process information at all processes.
 */
	p = &procs[3 * myrank];
	p[0] = colour;
	p[1] = key;
	p[2] = myrank;

	err = MPI_Allgather(p, 3, MPI_INT, procs, 3, MPI_INT, comm);
	if (err != MPI_SUCCESS) {
		free((char *) procs);
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMSPLIT, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMSPLIT);
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT, err));
	}
/*
 * Processes with undefined colour can stop here.
 */
	if (colour == MPI_UNDEFINED) {
		*pnewcomm = MPI_COMM_NULL;
		free((char *) procs);
		_LAM_TRACE(__lam_trcffend_(BLKMPICOMMSPLIT, -1, comm, 0, 0));
		__lam_resetfunc_(BLKMPICOMMSPLIT);
		return(MPI_SUCCESS);
	}
/*
 * Sort the process information.
 * Locate and count the # of processes having my colour.
 */
	qsort((char *) procs, nprocs, 3 * sizeof(int), cmpprocs);

	myprocs = 0;

	for (i = 0, p = procs; (i < nprocs) && (*p != colour); ++i, p += 3);

	myprocs = p;
	mynprocs = 1;

	for (++i, p += 3; (i < nprocs) && (*p == colour); ++i, p += 3) {
		++mynprocs;
	}
/*
 * Compact the old ranks of my old group in the array.
 */
	p = myprocs;
	p2 = myprocs + 2;

	for (i = 0; i < mynprocs; ++i, ++p, p2 += 3) {
		*p = *p2;
	}
/*
 * Create the new group.
 */
	MPI_Comm_group(comm, &oldgroup);

	err = MPI_Group_incl(oldgroup, mynprocs, myprocs, &newgroup);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT, err));
	}

	free((char *) procs);
	MPI_Group_free(&oldgroup);
/*
 * Create the new communicator.
 * MPI_Comm_create() is not called because we don't want to communicate
 * globally again, we have the new contextid (and the undefined-colour
 * procs have returned), so we win speed and safety.
 */
	newcomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (newcomm == 0) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pnewcomm = newcomm;
	newcomm->c_inter = 0;
	newcomm->c_refcount = 1;
	newcomm->c_cube_dim = i = __lam_cubedim_(newgroup->g_nprocs);
	newcomm->c_group = newgroup;
	newcomm->c_contextid = cid;
	newcomm->c_errhdl = comm->c_errhdl;
	++(comm->c_errhdl->eh_refcount);
	newcomm->c_keys = 0;
	newcomm->c_topo_type = MPI_UNDEFINED;
	newcomm->c_topo_nprocs = 0;
	newcomm->c_topo_ndims = 0;
	newcomm->c_topo_dims = 0;
	newcomm->c_topo_index = 0;
	newcomm->c_topo_nedges = 0;
	newcomm->c_topo_edges = 0;
	newcomm->c_status = 0;

	__lam_setcid_(cid);

	if (__lam_trcomm_(newcomm)) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	if (! al_insert(__lam_comms_, &newcomm)) {
		return(__lam_errfunc_(comm, BLKMPICOMMSPLIT,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}
/*
 * Generate run time trace.
 */
	_LAM_TRACE(__lam_trcffend_(BLKMPICOMMSPLIT, -1, comm, 0, 0));

	__lam_resetfunc_(BLKMPICOMMSPLIT);
	return(MPI_SUCCESS);
}

/*
 *	cmpprocs
 *
 *	Function:	- compare two process entries
 *	Accepts:	- ptr 1st entry
 *			- ptr 2nd entry
 *	Returns:	- -1, 0, +1 (less, equal, greater)
 */
static int
cmpprocs(e1, e2)

int			*e1, *e2;

{
	int		cmp;

	if (*e1 < *e2) {
		cmp = -1;
	} else if (*e1 > *e2) {
		cmp = 1;
	} else if (*(++e1) < *(++e2)) {
		cmp = -1;
	} else if (*e1 > *e2) {
		cmp = 1;
	} else if (*(++e1) < *(++e2)) {
		cmp = -1;
	} else if (*e1 > *e2) {
		cmp = 1;
	} else {
		cmp = 0;
	}

	return(cmp);
}
