/*
 *	Ohio Trollius
 *	Copyright 1994 The Ohio State University
 *	RBD
 *
 *	$Log:	dimscreate.c,v $
 * Revision 6.0  96/02/29  14:12:20  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:21:14  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:07  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:50:39  gdburns
 * Ohio Release
 * 
 * Revision 2.3  94/05/18  12:48:11  gdburns
 * new for 5.1
 * 
 *	Function:	- select "best" grid dimensions
 *			- "Greed is good, greed works!" (Gecko)
 *	Accepts:	- # of processes
 *			- # of dimensions
 *			- array of dimension sizes
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>
#include <typical.h>

/*
 * static functions
 */
static int		assignprocs();
static int		getfactors();
static int		getprimes();

int
MPI_Dims_create(nprocs, ndims, dims)

int			nprocs;
int			ndims;
int			*dims;

{
	int		i;
	int		freeprocs;
	int		freedims;
	int		nprimes;
	int		*primes;
	int		*factors;
	int		*procs;
	int		*p;

	__lam_initerr_();
	__lam_setfunc_(BLKMPIDIMSCREATE);
/*
 * Check the arguments.
 */
	if (ndims < 1) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if ((nprocs <= 0) || (dims == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Get # of free-to-be-assigned processes and # of free dimensions.
 */
	freeprocs = nprocs;
	freedims = 0;
	for (i = 0, p = dims; i < ndims; ++i,++p) {
		if (*p == 0) {
			++freedims;
		} else if ((*p < 0) || ((nprocs % *p) != 0)) {
			return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIDIMSCREATE,
					__lam_mkerr_(MPI_ERR_DIMS, 0)));
		} else {
			freeprocs /= *p;
		}
	}

	if (freedims == 0) {
		if (freeprocs == 1) {
			__lam_resetfunc_(BLKMPIDIMSCREATE);
			return(MPI_SUCCESS);
		}

		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIDIMSCREATE,
				__lam_mkerr_(MPI_ERR_DIMS, 0)));
	}

	if (freeprocs < 1) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_DIMS, 0)));
	}
	else if (freeprocs == 1) {
		for (i = 0; i < ndims; ++i, ++dims) {
			if (*dims == 0) {
				*dims = 1;
			}
		}

		__lam_resetfunc_(BLKMPIDIMSCREATE);
		return(MPI_SUCCESS);
	}
/*
 * Compute the relevant prime numbers for factoring.
 */
	if (getprimes(freeprocs, &nprimes, &primes)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Factor the number of free processes.
 */
	if (getfactors(freeprocs, nprimes, primes, &factors)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Assign free processes to free dimensions.
 */
	if (assignprocs(freedims, nprimes, primes, factors, &procs)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIDIMSCREATE, __lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Return assignment results.
 */
	p = procs;
	for (i = 0; i < ndims; ++i, ++dims) {
		if (*dims == 0) {
			*dims = *p++;
		}
	}

	free((char *) primes);
	free((char *) factors);
	free((char *) procs);

	__lam_resetfunc_(BLKMPIDIMSCREATE);
	return(MPI_SUCCESS);
}

/*
 *	assignprocs
 *
 *	Function:	- assign processes to dimensions
 *			- get "best-balanced" grid
 *			- greedy bin-packing algorithm used
 *			- sort dimensions in decreasing order
 *			- dimensions array dynamically allocated
 *	Accepts:	- # of dimensions
 *			- # of prime factors
 *			- array of prime factors
 *			- array of factor counts
 *			- ptr to array of dimensions (returned value)
 *	Returns:	- 0 or ERROR
 */
static int
assignprocs(ndim, nfactor, pfacts, counts, pdims)

int			ndim;
int			nfactor;
int			*pfacts;
int			*counts;
int			**pdims;

{
	int		*bins;
	int		i, j;
	int		n;
	int		f;
	int		*p;
	int		*pmin;

	if (ndim <= 0) {
		errno = EINVAL;
		return(ERROR);
	}
/*
 * Allocate and initialize the bins.
 */
	bins = (int *) malloc((unsigned) ndim * sizeof(int));
	if (bins == 0) {
		return(ERROR);
	}
	*pdims = bins;

	for (i = 0, p = bins; i < ndim; ++i, ++p) {
		*p = 1;
	}
/*
 * Loop assigning factors from the highest to the lowest.
 */
	for (j = nfactor - 1; j >= 0; --j) {

		f = pfacts[j];
		for (n = counts[j]; n > 0; --n) {
/*
 * Assign a factor to the smallest bin.
 */
			pmin = bins;
			for (i = 1, p = pmin + 1; i < ndim; ++i, ++p) {
				if (*p < *pmin) {
					pmin = p;
				}
			}

			*pmin *= f;
		}
	}
/*
 * Sort dimensions in decreasing order (O(n^2) for now).
 */
	for (i = 0, pmin = bins; i < ndim - 1; ++i, ++pmin) {
		for (j = i + 1, p = pmin + 1; j < ndim; ++j, ++p) {
			if (*p > *pmin) {
				n = *p;
				*p = *pmin;
				*pmin = n;
			}
		}
	}

	return(0);
}

/*
 *	getfactors
 *
 *	Function:	- factorize a number
 *	Accepts:	- number
 *			- # of primes
 *			- array of primes
 *			- ptr to array of counts (returned value)
 *	Returns:	- 0 or ERROR
 */
static int
getfactors(num, nprime, primes, pcounts)

int			num;
int			nprime;
int			*primes;
int			**pcounts;

{
	int		*counts;
	int		i;
	int		*p;
	int		*c;

	if (nprime <= 0) {
		errno = EINVAL;
		return(ERROR);
	}
/*
 * Allocate the factor counts array.
 */
	counts = (int *) malloc((unsigned) nprime * sizeof(int));
	if (counts == 0) {
		return(ERROR);
	}
	*pcounts = counts;
/*
 * Loop over the prime numbers.
 */
	i = nprime - 1;
	p = primes + i;
	c = counts + i;

	for (; i >= 0; --i, --p, --c) {
		*c = 0;
		while ((num % *p) == 0) {
			++(*c);
			num /= *p;
		}
	}

	if (num != 1) {
		errno = EIMPOSSIBLE;
		return(ERROR);
	}

	return(0);
}

/*
 *	getprimes
 *
 *	Function:	- get primes smaller than number
 *			- array of primes dynamically allocated
 *	Accepts:	- number
 *			- ptr # of primes (returned value)
 *			- ptr array of primes (returned values)
 *	Returns:	- 0 or ERROR
 */
static int
getprimes(num, pnprime, pprimes)

int			num;
int			*pnprime;
int			**pprimes;

{
	int		i, j;
	int		n;
	int		size;
	int		*primes;
/*
 * Allocate the array of primes.
 */
	size = (num / 2) + 1;
	primes = (int *) malloc((unsigned) size * sizeof(int));
	if (primes == 0) {
		return(ERROR);
	}
	*pprimes = primes;
/*
 * Find the prime numbers.
 */
	i = 0;
	primes[i++] = 2;

	for (n = 3; n <= num; n += 2) {
		for (j = 1; j < i; ++j) {
			if ((n % primes[j]) == 0) {
				break;
			}
		}

		if (j == i) {
			if (i >= size) {
				errno = EIMPOSSIBLE;
				return(ERROR);
			}

			primes[i++] = n;
		}
	}

	*pnprime = i;
	return(0);
}
