/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	finalize.c,v $
 * Revision 6.0  96/02/29  14:12:26  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/28  23:37:05  nevin
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:14  gdburns
 * Ohio Release
 * 
 * Revision 5.1.1.1  94/07/28  15:52:17  raja
 * Change _kio_cleanup() to _cio_cleanup().
 * 
 * Revision 5.1  94/05/18  12:50:45  gdburns
 * Ohio Release
 * 
 *	Function:	- terminate the MPI session
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>
#include <typical.h>

/*
 * external functions
 */
extern void		_cio_cleanup();
extern void		__lam_initerr_();
extern void		__lam_nukecids_();
extern void		__lam_nukefunc_();
extern void		__lam_nukekeys_();
extern void		__lam_nukeprocs_();
extern void		__lam_nuketrace_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trcffstart_();
extern void             _mpi_req_get();
extern void		lam_kexit();
extern int		__lam_errfunc_();
extern int              _mpi_req_getn();
extern int		_rpi_c2c_finalize();
extern int		_rpi_lamd_finalize();

/*
 * local functions
 */
static int		waitall();
static void		free_comm();
static void		free_dtype();
static void		free_env();
static void		free_errhdl();
static void		free_op();
static void		free_rdtype();

int
MPI_Finalize()

{
	int		err;			/* error code */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIFINAL);

	_LAM_TRACE(__lam_trcffstart_(BLKMPIFINAL));
/*
 * Block until all pending requests are done.
 */
	err = waitall();
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIFINAL, err));
	}
/*
 * RPI specific finalization.
 */
	if (RPI_SPLIT(_rpi_lamd_finalize, _rpi_c2c_finalize, ())) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIFINAL,
					__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}
/*
 * Cleanup.
 */
	free_env();
	free_rdtype();
	free_comm();
	free_op();
	free_dtype();
	free_errhdl();

	_LAM_TRACE(__lam_trcffend_(BLKMPIFINAL, -1, 0, 0, 0));

	__lam_troff_();
	__lam_nuketrace_();
 	__lam_nukeprocs_();
	lam_kexit(0);

	__lam_resetfunc_(BLKMPIFINAL);
	__lam_nukefunc_();
	
	return(MPI_SUCCESS);
}

/*
 *	waitall
 *
 *	Function:	- wait till all pending requests are done
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
waitall()

{
	int		err;			/* error code */
	int		nreqs;			/* # pending requests */
	MPI_Request	*reqs;			/* request array */
	MPI_Status	*stats;			/* status array */

	nreqs = _mpi_req_getn();
	if (nreqs == 0) return(MPI_SUCCESS);

	reqs = (MPI_Request *) malloc((unsigned) nreqs * sizeof(MPI_Request));
	stats = (MPI_Status *) malloc((unsigned) nreqs * sizeof(MPI_Status));

	if ((reqs == 0) || (stats == 0)) {
		if (reqs) free((char *) reqs);
		if (stats) free((char *) stats);
		return(__lam_mkerr_(MPI_ERR_OTHER, errno));
	}

	_mpi_req_get(nreqs, reqs);

	err = MPI_Waitall(nreqs, reqs, stats);

	free((char *) reqs);
	free((char *) stats);

	return(err);
}

/*
 *	free_errhdl
 *
 *	Function:	- deallocate pre-defined error handles
 */
static void
free_errhdl()

{
/*
 * Empty stub, nothing to free.
 */
}

/*
 *	free_comm
 *
 *	Function:	- deallocate pre-defined communicators
 */
static void
free_comm()

{
/*
 * Free the "world" communicator.
 */
	__lam_rmcid_(MPI_COMM_WORLD->c_contextid);
	free((char *) MPI_COMM_WORLD->c_group);
/*
 * Free the "self" communicator.
 */
	__lam_rmcid_(MPI_COMM_SELF->c_contextid);
	free(MPI_COMM_SELF->c_group);

	__lam_nukecids_();
}

/*
 *	free_dtype
 *
 *	Function:	- deallocate basic (intrinsic) datatypes
 */
static void
free_dtype()

{
/*
 * Empty stub, nothing to free.
 */
}

/*
 *	free_rdtype
 *
 *	Function:	- deallocate reduction datatypes
 */
static void
free_rdtype()

{
	MPI_Datatype	dtype;			/* favourite datatype */

	dtype = MPI_2INT; MPI_Type_free(&dtype);
	dtype = MPI_2FLOAT; MPI_Type_free(&dtype);
	dtype = MPI_2DOUBLE; MPI_Type_free(&dtype);
	dtype = MPI_FLOAT_INT; MPI_Type_free(&dtype);
	dtype = MPI_DOUBLE_INT; MPI_Type_free(&dtype);
	dtype = MPI_LONG_INT; MPI_Type_free(&dtype);
}

/*
 *	free_op
 *
 *	Function:	- deallocate intrinsic reduction operations
 */
static void
free_op()

{
/*
 * Empty stub, nothing to free.
 */
}

/*
 *	free_env
 *
 *	Function:	- deallocate environment attributes
 */
static void
free_env()

{
	int		key;			/* favourite key */

	MPI_Attr_delete(MPI_COMM_WORLD, MPI_TAG_UB);
	MPI_Attr_delete(MPI_COMM_WORLD, MPI_HOST);
	MPI_Attr_delete(MPI_COMM_WORLD, MPI_IO);

	key = MPI_TAG_UB; MPI_Keyval_free(&key);
	key = MPI_HOST; MPI_Keyval_free(&key);
	key = MPI_IO; MPI_Keyval_free(&key);

	__lam_nukekeys_();
}
