/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	icmerge.c,v $
 * Revision 6.0  96/02/29  14:12:50  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.1  96/02/18  21:22:56  gdburns
 * major overhaul
 * 
 * Revision 5.2  94/08/22  14:04:31  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:04  gdburns
 * Ohio Release
 * 
 *	Function:	- merge inter-communicator into intra-communicator
 *	Accepts:	- inter-communicator
 *			- high/low flag
 *			- ptr new intra-communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Intercomm_merge(comm, high, pnewcomm)

MPI_Comm		comm;
int			high;
MPI_Comm		*pnewcomm;

{
	MPI_Comm	newcomm;		/* new communicator */
	MPI_Group	newgroup;		/* new group */
	MPI_Group	g1, g2;			/* 2 groups to merge */
	MPI_Status	stat;			/* message status */
	int		mycid;			/* my context ID */
	int		cid;			/* global context ID */
	int		rhigh;			/* remote high value */
	int		err;			/* error code */
	int		inmsg[2];		/* input message buffer */
	int		outmsg[2];		/* output message buffer */
	int		i;			/* favourite index */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIICOMMMERGE);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIICOMMMERGE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (comm->c_inter == 0) {
		return(__lam_errfunc_(comm,
			BLKMPIICOMMMERGE, __lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (pnewcomm == 0) {
		return(__lam_errfunc_(comm,
			BLKMPIICOMMMERGE, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	_LAM_TRACE(__lam_trcffstart_(BLKMPIICOMMMERGE));
/*
 * Create the new context ID: rank0 processes are the leaders of their
 * local group.  Each leader finds the max context ID of all remote group
 * processes (excluding their leader) and their "high" setting.
 * The leaders then swap the information and broadcast to the remote group.
 * Note: this is a criss-cross effect, processes talk to the remote leader.
 */
	mycid = __lam_getcid_();
	if (mycid < 0) {
		return(__lam_errfunc_(comm, BLKMPIICOMMMERGE,
				__lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}

	if (comm->c_group->g_myrank == 0) {
		cid = mycid;
/*
 * Find max context ID of remote non-leader processes.
 */
		for (i = 1; i < comm->c_rgroup->g_nprocs; ++i) {

			err = MPI_Recv(inmsg, 1, MPI_INT,
						i, MPI_MERGE_TAG, comm, &stat);
			if (err != MPI_SUCCESS) {
				return(__lam_errfunc_(comm,
						BLKMPIICOMMMERGE, err));
			}

			if (inmsg[0] > cid) cid = inmsg[0];
		}
/*
 * Swap context ID and high value with remote leader.
 */
		outmsg[0] = cid;
		outmsg[1] = high;

		err = MPI_Sendrecv(outmsg, 2, MPI_INT, 0, MPI_MERGE_TAG,
					inmsg, 2, MPI_INT, 0, MPI_MERGE_TAG,
					comm, &stat);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIICOMMMERGE, err));
		}

		if (inmsg[0] > cid) cid = inmsg[0];

		rhigh = inmsg[1];
/*
 * Broadcast context ID and local high to remote non-leader processes.
 */
		outmsg[0] = cid;
		outmsg[1] = high;

		for (i = 1; i < comm->c_rgroup->g_nprocs; ++i) {

			err = MPI_Send(outmsg, 2, MPI_INT,
						i, MPI_MERGE_TAG, comm);
			if (err != MPI_SUCCESS) {
				return(__lam_errfunc_(comm,
						BLKMPIICOMMMERGE, err));
			}
		}
	}
/*
 * non-leader
 */
	else {
		err = MPI_Send(&mycid, 1, MPI_INT, 0, MPI_MERGE_TAG, comm);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIICOMMMERGE, err));
		}

		err = MPI_Recv(inmsg, 2, MPI_INT, 0,
						MPI_MERGE_TAG, comm, &stat);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIICOMMMERGE, err));
		}

		cid = inmsg[0];
		rhigh = inmsg[1];
	}
/*
 * All procs know the "high" for local and remote groups and the context ID.
 * Create the properly ordered union group.  In case of equal high values,
 * the group that has the leader with the "lowest process" goes first.
 */
	if (high && (!rhigh)) {
		g1 = comm->c_rgroup;
		g2 = comm->c_group;
	} else if ((!high) && rhigh) {
		g1 = comm->c_group;
		g2 = comm->c_rgroup;
	} else if (__lam_proccmp_(comm->c_group->g_procs[0],
				comm->c_rgroup->g_procs[0]) < 0) {
		g1 = comm->c_group;
		g2 = comm->c_rgroup;
	} else {
		g1 = comm->c_rgroup;
		g2 = comm->c_group;
	}

	err = MPI_Group_union(g1, g2, &newgroup);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIICOMMMERGE, err));
	}
/*
 * Create the new communicator.
 */
	newcomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (newcomm == 0) {
		return(__lam_errfunc_(comm, BLKMPIICOMMMERGE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*pnewcomm = newcomm;
	newcomm->c_inter = 0;
	newcomm->c_refcount = 1;
	newcomm->c_cube_dim = __lam_cubedim_(newgroup->g_nprocs);
	newcomm->c_group = newgroup;
	newcomm->c_rgroup = MPI_GROUP_NULL;
	newcomm->c_contextid = cid;
	newcomm->c_errhdl = comm->c_errhdl;
	++(comm->c_errhdl->eh_refcount);
	newcomm->c_keys = 0;
	newcomm->c_topo_type = MPI_UNDEFINED;
	newcomm->c_topo_nprocs = 0;
	newcomm->c_topo_ndims = 0;
	newcomm->c_topo_dims = 0;
	newcomm->c_topo_index = 0;
	newcomm->c_topo_nedges = 0;
	newcomm->c_topo_edges = 0;
	newcomm->c_status = 0;

	if (! al_insert(__lam_comms_, &newcomm)) {
		return(__lam_errfunc_(comm, BLKMPIICOMMMERGE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	if (__lam_trcomm_(newcomm)) {
		return(__lam_errfunc_(comm, BLKMPIICOMMMERGE,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	__lam_setcid_(cid);

	_LAM_TRACE(__lam_trcffend_(BLKMPIICOMMMERGE, -1, comm, 0, 0));

	__lam_resetfunc_(BLKMPIICOMMMERGE);
	return(MPI_SUCCESS);
}
