/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	lambuf.c,v $
 * Revision 6.0  96/02/29  14:13:00  gdburns
 * Ohio Release
 * 
 *	Function:	- buffer management functions
 */

#include <stdlib.h>
#include <errno.h>

#include <blktype.h>
#include <mpi.h>
#include <t_types.h>
#include <terror.h>

/*
 * macros
 */
#define HDRSIZE		(sizeof(struct _bsndhdr))

/*
 * public functions
 */
int			__lam_bufinit_bsend_();
int			__lam_bufinit_();
void			__lam_buflength_();
int			__lam_buffree_bsend_();
int			__lam_bufattach_();
int			__lam_bufdetach_();

/*
 * external functions
 */
extern int		__lam_dtflag_();
extern int		__lam_mkerr_();
extern int		_mpi_req_advance();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();

/*
 * local variables
 */
static void		*userbuf = 0;
static int		bufsize = 0;
static struct _bsndhdr	*buftop = 0;
static struct _bsndhdr	*bufend = 0;
static struct _bsndhdr	*buffree = 0;

/*
 * local functions
 */
static void		buf_append();
static void		buf_free();
static void		buf_unlink();
static struct _bsndhdr	*buf_alloc();

/*
 *	__lam_bufinit_bsend_
 *
 *	Function:	- setup bsend buffer
 *	Accepts:	- request
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_bufinit_bsend_(rq)

MPI_Request		rq;

{
/*
 * Handle zero length messages.
 */
	if ((rq->rq_count == 0) || (rq->rq_dtype->dt_size == 0)) {
		rq->rq_packbuf = (char *) rq->rq_buf;
		rq->rq_packsize = 0;
		return(MPI_SUCCESS);
	}

	rq->rq_packsize = rq->rq_count * rq->rq_dtype->dt_size;
/*
 * Allocate a chunk of buffer for the message.
 */
	rq->rq_bsend = buf_alloc(rq->rq_packsize);
	if (rq->rq_bsend == 0) return(__lam_mkerr_(MPI_ERR_OTHER, ENOBUFS));

	rq->rq_bsend->bsh_req = rq;

	buf_append(rq->rq_bsend);
/*
 * Setup the request for a typical packing buffer.
 * The message will be packed in and out by _mpi_req_start() and
 * _mpi_req_end().  We should optimize with a memcpy() and Trollius
 * data conversion if the message is homogenous and contiguous.
 */
	rq->rq_packbuf = ((char *) rq->rq_bsend) + HDRSIZE;
	rq->rq_marks |= LAM_RQFCHAR;

	return(MPI_SUCCESS);
}

/*
 *	__lam_bufinit_
 *
 *	Function:	- setup send/recv message buffer
 *	Accepts:	- request
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_bufinit_(rq)

MPI_Request		rq;

{
	int		fl_contig;		/* contig. buffer flag */
	MPI_Datatype	dtype;			/* datatype */
	double		local_rep;		/* local data representation */
	double		net_rep;		/* net data representation */

	dtype = rq->rq_dtype;
/*
 * Handle zero length messages.
 */
	if ((rq->rq_count == 0) || (dtype->dt_size == 0)) {
		rq->rq_packbuf = (char *) rq->rq_buf;
		rq->rq_packsize = 0;
		return(MPI_SUCCESS);
	}

	fl_contig = (dtype->dt_format == LAM_DTBASIC) ||
			((dtype->dt_format == LAM_DTCONTIG) &&
				(dtype->dt_dtype->dt_format == LAM_DTBASIC));
	local_rep = 1.1;
	ltotf8(&local_rep, &net_rep);
/*
 * If contiguous, use the caller's buffer.
 */
	rq->rq_packsize = rq->rq_count * dtype->dt_size;

	if (fl_contig && ((local_rep == net_rep) || __lam_homog_)) {
		rq->rq_packbuf = (char *) rq->rq_buf;
/*
 *		rq->rq_marks |= __lam_dtflag_(dtype);
 */
		rq->rq_marks |= LAM_RQFCHAR;
	}
/*
 * Otherwise allocate a buffer.
 */
	else {
		rq->rq_packbuf = malloc((unsigned) rq->rq_packsize);
		if (rq->rq_packbuf == 0) {
			return(__lam_mkerr_(MPI_ERR_OTHER, errno));
		}

		rq->rq_marks |= LAM_RQFCHAR;
	}

	return(MPI_SUCCESS);
}

/*
 *	__lam_buflength_
 *
 *	Function:	- translates message length into count and #elements
 *	Accepts:	- request
 */
void
__lam_buflength_(rq)

MPI_Request		rq;

{
	int		size;			/* datatype size */
	int		fl_contig;		/* contig. buffer flag */
	MPI_Datatype	dtype;			/* datatype */

	dtype = rq->rq_dtype;
/*
 * For a datatype of zero size, count and #elements are undefined.
 */
	if (dtype->dt_size == 0) {
		rq->rq_status.st_count = MPI_UNDEFINED;
		rq->rq_status.st_nelem = MPI_UNDEFINED;
		return;
	}
	
	fl_contig = (dtype->dt_format == LAM_DTBASIC) ||
			((dtype->dt_format == LAM_DTCONTIG) &&
				(dtype->dt_dtype->dt_format == LAM_DTBASIC));
/*
 * If contiguous, set the #elements.
 */
	if (fl_contig) {
		size = (dtype->dt_format == LAM_DTBASIC) ?
				dtype->dt_size : dtype->dt_dtype->dt_size;
		rq->rq_status.st_nelem = rq->rq_status.st_length / size;
	}
/*
 * If not contiguous, defer the computation of #elements.
 */
	else {
		rq->rq_status.st_nelem = -1;
	}
/*
 * Compute the datatype count.  A non-integral count becomes undefined.
 */
	rq->rq_status.st_count =
			((rq->rq_status.st_length % dtype->dt_size) != 0) ?
			MPI_UNDEFINED : rq->rq_status.st_length /
			dtype->dt_size;
}

/*
 *	__lam_buffree_bsend_
 *
 *	Function:	- free bsend buffer
 *	Accepts:	- request
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_buffree_bsend_(req)

MPI_Request		req;

{
/*
 * Handle zero length messages.
 */
	if ((req->rq_count == 0) || (req->rq_dtype->dt_size == 0)) {
		return(MPI_SUCCESS);
	}

	buf_unlink(req->rq_bsend);
	buf_free(req->rq_bsend);
	req->rq_packbuf = 0;

	return(MPI_SUCCESS);
}

/*
 *	__lam_bufattach_
 *
 *	Function:	- set the user buffer space
 *	Accepts:	- buffer
 *			- buffer size
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_bufattach_(buf, size)

void			*buf;
int			size;

{
	int		err;			/* error code */

	if ((userbuf == 0) && (size >= sizeof(struct _bsndhdr))) {

		userbuf = buf;
		bufsize = size;

		size /= sizeof(void *);
		size *= sizeof(void *);

		buffree = (struct _bsndhdr *) buf;
		buffree->bsh_req = 0;
		buffree->bsh_size = size - sizeof(struct _bsndhdr);
		buffree->bsh_prev = 0;
		buffree->bsh_next = 0;

		err = MPI_SUCCESS;
	}
	else {
		err = __lam_mkerr_(MPI_ERR_OTHER, (userbuf) ? EBUSY : EINVAL);
	}

	return(err);
}

/*
 *	__lam_bufdetach_
 *
 *	Function:	- detach and return user buffer
 *			- block until all requests are done
 *	Accepts:	- ptr buffer
 *			- ptr buffer size
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_bufdetach_(pbuf, psize)

void			*pbuf;
int			*psize;

{
	int		err;			/* error code */
	struct _bsndhdr	*p;			/* favourite pointer */
/*
 * Block until all bsend requests are done.
 * Since all bsend requests are orphans, they will be cleaned up
 * by _mpi_req_advance().
 */
	while (buftop) {
/*
 * Mark all bsend requests as blocking and advance the system.
 */
		_mpi_req_blkclr();

		for (p = buftop; p; p = p->bsh_next) {
			_mpi_req_blkset(p->bsh_req);
		}

		err = _mpi_req_advance();
		if (err != MPI_SUCCESS) return(err);
	}
/*
 * Return the buffer to the user.
 */
	*((void **) pbuf) = userbuf;
	*psize = bufsize;

	userbuf = 0;
	bufsize = 0;
	buftop = bufend = buffree = 0;
 
	return(MPI_SUCCESS);
}

/*
 *	buf_alloc
 *
 *	Function:	- allocate a chunk of buffer
 *			- sizes are forced to multiple of pointer size
 *	Accepts:	- buffer size
 *	Returns:	- ptr allocated buffer header or NULL
 */
static struct _bsndhdr *
buf_alloc(size)

int			size;

{
	struct _bsndhdr	*pchunk;		/* ptr allocated chunk */
	struct _bsndhdr	*pcur;			/* ptr current block */
	struct _bsndhdr	*pprev;			/* ptr previous block */
/*
 * Round up the size to a multiple of pointer size.
 */
	size = (size + sizeof(void *) - 1) / sizeof(void *);
	size *= sizeof(void *);
/*
 * Loop searching for a big-enough free chunk.
 */
	for (pcur = buffree, pprev = 0; pcur; pcur = pcur->bsh_next) {
		if (pcur->bsh_size >= size) break;
		pprev = pcur;
	}

	if (pcur == 0) return((struct _bsndhdr *) 0);
/*
 * Take a slice from the free chunk.
 */
	if ((pcur->bsh_size - size) >= (2 * HDRSIZE)) {
		pchunk = (struct _bsndhdr *)
				(((char *) pcur) + pcur->bsh_size - size);
		pchunk->bsh_size = size;
		pcur->bsh_size -= size + HDRSIZE;
	}
/*
 * Chunk not worth slicing, take it all.
 */
	else {
		pchunk = pcur;

		if (pprev == 0) {
			buffree = pcur->bsh_next;
		} else {
			pprev->bsh_next = pcur->bsh_next;
		}
	}

	pchunk->bsh_req = 0;
	pchunk->bsh_next = 0;
	pchunk->bsh_prev = 0;

	return(pchunk);
}

/*
 *	buf_free
 *
 *	Function:	- free a chunk of buffer
 *			- add it to the sorted free list
 *			- coalesce with up/down neighbours if possible
 *	Accepts:	- ptr chunk
 */
static void
buf_free(pchunk)

struct _bsndhdr		*pchunk;

{
	struct _bsndhdr	*pcur;			/* ptr current chunk */
	struct _bsndhdr	*pprev;			/* ptr previous chunk */
/*
 * Locate the chunk's place in the free list.
 */
	for (pcur = buffree, pprev = 0; pcur; pcur = pcur->bsh_next) {
		if (((long) pcur) > ((long) pchunk)) break;
		pprev = pcur;
	}
/*
 * Link the chunk in the list.
 */
	if (pprev == 0) {
		buffree = pchunk;
	} else {
		pprev->bsh_next = pchunk;
	}

	pchunk->bsh_next = pcur;
/*
 * Coalesce with previous if possible.
 */
	if (pprev) {
		if ( ((long) pchunk) ==
			(((long) pprev) + HDRSIZE + pprev->bsh_size) ) {

			pprev->bsh_next = pchunk->bsh_next;
			pprev->bsh_size += HDRSIZE + pchunk->bsh_size;
			pchunk = pprev;
		}
	}
/*
 * Coalesce with next if possible.
 */
	if (pcur) {
		if (((long) pcur) ==
			(((long) pchunk) + HDRSIZE + pchunk->bsh_size) ) {

			pchunk->bsh_next = pcur->bsh_next;
			pchunk->bsh_size += HDRSIZE + pcur->bsh_size;
		}
	}
}

/*
 *	buf_append
 *
 *	Function:	- append buffer to list of requests
 *	Accepts:	- ptr buffer
 */
static void
buf_append(pbuf)

struct _bsndhdr		*pbuf;

{
	pbuf->bsh_prev = bufend;
	pbuf->bsh_next = 0;

	if (bufend == 0) buftop = pbuf;
	else bufend->bsh_next = pbuf;

	bufend = pbuf;
}

/*
 *	buf_unlink
 *
 *	Function:	- unlink buffer from list of requests
 *	Accepts:	- ptr buffer entry
 */
static void
buf_unlink(pbuf)

struct _bsndhdr		*pbuf;

{
	struct _bsndhdr	*pprev;			/* ptr previous buffer */
	struct _bsndhdr	*pnext;			/* ptr next buffer */

	pprev = pbuf->bsh_prev;
	pnext = pbuf->bsh_next;

	if (pprev == 0) buftop = pnext;
	else pprev->bsh_next = pnext;

	if (pnext) pnext->bsh_prev = pprev;
	else bufend = pprev;
}
