/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	lamdtype.c,v $
 * Revision 6.0  96/02/29  14:13:03  gdburns
 * Ohio Release
 * 
 *	Function:	- misc. datatype functions
 */

#include <mpi.h>

/*
 * global functions
 */
int			__lam_dtflag_();
int			__lam_dtsndrcv_();
void			__lam_dtcpy_();

/*
 * local functions
 */
static void		cpy_hvector();
static void		cpy_hindexed();
static void		cpy_struct();

/*
 *	__lam_dtflag_
 *
 *	Function:	- get the basic datatype flag
 *	Accepts:	- datatype
 *	Returns:	- flag
 */
int
__lam_dtflag_(type)

MPI_Datatype		type;

{
	int		flag = 0;		/* datatype flag */

	if ((type == MPI_CHAR) || (type == MPI_UNSIGNED_CHAR) ||
			(type == MPI_BYTE)) {
		flag = LAM_RQFCHAR;
	}
	else if ((type == MPI_INT) || (type == MPI_UNSIGNED) ||
		(type == MPI_LONG) || (type == MPI_UNSIGNED_LONG) ||
					(type == MPI_LOGIC)) {
		flag = LAM_RQFINT;
	}
	else if ((type == MPI_FLOAT) || (type == MPI_CPLEX)) {
		flag = LAM_RQFFLOAT;
	}
	else if ((type == MPI_DOUBLE) || (type == MPI_LONG_DOUBLE)) {
		flag = LAM_RQFDOUBLE;
	}
	else if ((type == MPI_SHORT) || (type == MPI_UNSIGNED_SHORT)) {
		flag = LAM_RQFSHORT;
	}

	return(flag);
}

/*
 *	__lam_dtsndrcv_
 *
 *	Function:	- copy MPI message from buffer into another
 *			- send/recv done if cannot optimize
 *	Accepts:	- send buffer
 *			- send count
 *			- send datatype
 *			- receive buffer
 *			- receive count
 *			- receive datatype
 *			- tag
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */
int
__lam_dtsndrcv_(sbuf, scount, sdtype, rbuf, rcount, rdtype, tag, comm)

void			*sbuf;
int			scount;
MPI_Datatype		sdtype;
void			*rbuf;
int			rcount;
MPI_Datatype		rdtype;
int			tag;
MPI_Comm		comm;

{
	int		err;			/* error code */
	int		size;			/* packed size */
	int		rank;			/* caller's rank */
	MPI_Status	stat;			/* status info */

	err = __lam_mkerr_(MPI_ERR_TRUNCATE, 0);
/*
 * If same datatypes used, just copy.
 */
	if (sdtype == rdtype) {
		if (scount <= rcount) {
			__lam_dtcpy_((char *) rbuf, (char *) sbuf, rcount,
					rdtype);
			err = MPI_SUCCESS;
		}
	}
/*
 * If receive packed.
 */
	else if (rdtype == MPI_PACKED) {
		MPI_Pack_size(scount, sdtype, MPI_COMM_WORLD, &size);
		if (size <= rcount) {
			if (__lam_pack_(sbuf, scount, sdtype,
						rbuf, rcount) == size) {
				err = MPI_SUCCESS;
			}
		}
	}
/*
 * If send packed.
 */
	else if (sdtype == MPI_PACKED) {
		MPI_Pack_size(rcount, rdtype, MPI_COMM_WORLD, &size);
		if (size >= scount) {
			if (__lam_unpack_(sbuf, scount,
					rbuf, rcount, rdtype) == scount) {
				err = MPI_SUCCESS;
			}
		}
	}
/*
 * Let send/recv handle it.
 */
	else {
		MPI_Comm_rank(comm, &rank);
		err = MPI_Sendrecv(sbuf, scount, sdtype, rank, tag,
				rbuf, rcount, rdtype, rank, tag, comm, &stat);
	}

	return(err);
}

/*
 *	__lam_dtcpy_
 *
 *	Function:	- copy a datatype from a buffer into another
 *			- copying is done recursively
 *	Accepts:	- destination buffer
 *			- source buffer
 *			- count
 *			- datatype
 */
void
__lam_dtcpy_(destbuf, srcbuf, count, dtype)

char			*destbuf;
char			*srcbuf;
int			count;
MPI_Datatype		dtype;

{
/*
 * Handle the different datatypes.
 */
	switch(dtype->dt_format) {

	case LAM_DTBASIC:
		if ((dtype == MPI_UNSIGNED_CHAR) || (dtype == MPI_PACKED) ||
				(dtype == MPI_BYTE) || (dtype == MPI_CHAR)) {

			memcpy(destbuf, srcbuf, count);

		}
		else if ((dtype == MPI_INT) || (dtype == MPI_UNSIGNED) ||
				(dtype == MPI_LOGIC)) {

			memcpy(destbuf, srcbuf, count * sizeof(int));
		}
		else if ((dtype == MPI_LONG) ||
				(dtype == MPI_UNSIGNED_LONG)) {

			memcpy(destbuf, srcbuf, count * sizeof(long));
		}
		else if ((dtype == MPI_SHORT) ||
				(dtype == MPI_UNSIGNED_SHORT)) {

			memcpy(destbuf, srcbuf, count * sizeof(short));
		}
		else if (dtype == MPI_FLOAT) {

			memcpy(destbuf, srcbuf, count * sizeof(float));
		}
		else if (dtype == MPI_DOUBLE) {

			memcpy(destbuf, srcbuf, count * sizeof(double));
		}
		else if (dtype == MPI_LONG_DOUBLE) {

			memcpy(destbuf, srcbuf, 2 * count * sizeof(double));
		}
		else if (dtype == MPI_CPLEX) {

			memcpy(destbuf, srcbuf, 2 * count * sizeof(float));
		}

		break;

	case LAM_DTCONTIG:
		__lam_dtcpy_(destbuf, srcbuf,
				count * dtype->dt_count, dtype->dt_dtype);
		break;

	case LAM_DTVECTOR:
		cpy_hvector(destbuf, srcbuf, count, dtype, 1);
		break;

	case LAM_DTHVECTOR:
		cpy_hvector(destbuf, srcbuf, count, dtype, 0);
		break;

	case LAM_DTINDEXED:
		cpy_hindexed(destbuf, srcbuf, count, dtype, 1);
		break;

	case LAM_DTHINDEXED:
		cpy_hindexed(destbuf, srcbuf, count, dtype, 0);
		break;

	case LAM_DTSTRUCT:
		cpy_struct(destbuf, srcbuf, count, dtype);
		break;
	}
}

/*
 *	cpy_hvector
 *
 *	Function:	- copy vector or hvector
 */
static void
cpy_hvector(dest, src, num, dtype, fl_vec)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;
int			fl_vec;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		stride;			/* stride in bytes */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;
	stride = dtype->dt_stride;
	if (fl_vec) stride *= dtype->dt_dtype->dt_upper -
				dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		for (j = 0; j < dtype->dt_count; ++j) {

			__lam_dtcpy_(dbuf, sbuf,
					dtype->dt_length, dtype->dt_dtype);

			sbuf += stride;
			dbuf += stride;
		}
	}
}

/*
 *	cpy_hindexed
 *
 *	Function:	- copy indexed or hindexed
 */
static void
cpy_hindexed(dest, src, num, dtype, fl_idx)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;
int			fl_idx;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		factor;			/* extent factor */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;
	factor = (fl_idx == 0) ? 1 :
			dtype->dt_dtype->dt_upper - dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp) {

			__lam_dtcpy_(dbuf + (*disp * factor),
					sbuf + (*disp * factor),
					*len, dtype->dt_dtype);
		}
	}
}

/*
 *	cpy_struct
 *
 *	Function:	- copy struct
 */
static void
cpy_struct(dest, src, num, dtype)

char			*dest;
char			*src;
int			num;
MPI_Datatype		dtype;

{
	int		i, j;			/* favourite indices */
	int		extent;			/* datatype extent */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	MPI_Datatype	*type;			/* ptr datatypes */
	char		*sbuf, *dbuf;		/* temp. buffers */

	extent = dtype->dt_upper - dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		sbuf = src;
		dbuf = dest;

		src += extent;
		dest += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;
		type = dtype->dt_dtypes;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp, ++type) {

			__lam_dtcpy_(dbuf + *disp, sbuf + *disp, *len, *type);
		}
	}
}
