/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	NJN
 *
 *	$Log:	mpil_spawn.c,v $
 * Revision 6.0  96/02/29  14:13:20  gdburns
 * Ohio Release
 * 
 *	Function:	- spawn MPI processes
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <app_schema.h>
#include <blktype.h>
#include <events.h>
#include <mpi.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
int			MPIL_Comm_parent();
int			MPIL_Spawn();
int			MPIL_Universe_size();

/*
 * external variables
 */
extern int		argtailc;		/* from do_args() */
extern char		**argtailv;		/* from do_args() */

/*
 * local functions
 */
static int		spawn();

/*
 *	MPIL_Spawn
 *
 *	Function:	- spawn MPI processes
 *	Accepts:	- application schema
 *			- root process rank
 *			- parent communicator
 *			- ptr intercomm between parents and children (returned)
 *	Returns:	- MPI_SUCCESS or error code
 */
int
MPIL_Spawn(comm, schema, root, newcomm)

MPI_Comm 		comm;
char			*schema;
int			root;
MPI_Comm 		*newcomm;

{
	int		i;			/* favourite index */
	int		rank;			/* caller rank */
	int		size;			/* group size */
	int		err;			/* error code */
	int		numkids;		/* num. of children spawned */
	int		mycid;			/* local max context ID */
	int		cid;			/* context ID for intercomm */
	int		msg[2];			/* two int message buffer */
	struct _gps	*kids;			/* array of child GPS */
	struct _gps	*p;			/* favourite pointer */
	struct _proc	**pp;			/* another pointer */
	MPI_Comm	icomm;			/* parent-child intercomm. */
	MPI_Group	cg;			/* child group */
	MPI_Status	stat;
	
	__lam_initerr_();
	__lam_setfunc_(BLKMPILSPAWN);
/*
 * Check the arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	MPI_Comm_size(comm, &size);

	if ((root >= size) || (root < 0)) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_ROOT, 0)));
	}

	if ((newcomm == 0) || (schema == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPILSPAWN, __lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	_LAM_TRACE(__lam_trcffstart_(BLKMPILSPAWN));
/*
 * Set debugging parameters.
 */
	p = &(comm->c_group->g_procs[root]->p_gps);

	__lam_setparam_(BLKMPILSPAWN, root | (p->gps_grank << 16),
				(p->gps_node << 16) | p->gps_idx);
/*
 * Synchronize all members of the parent group and get the context ID
 * for the parent-child intercommunicator.
 */
	MPI_Comm_rank(comm, &rank);

	mycid = __lam_getcid_();

	if (mycid < 0) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_INTERN, EFULL)));
	}

	err = MPI_Reduce(&mycid, &cid, 1, MPI_INT, MPI_MAX, root, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN, err));
	}
	
	if (rank == root) {
/*
 * The root does the process spawning.
 */
		if (spawn(schema, comm, cid, &numkids, &kids)) {
/*
 * Inform parent group of spawn error.
 */
			msg[0] = -1; msg[1] = err;
			
			MPI_Bcast(msg, 2, MPI_INT, root, comm);
			
			return(__lam_errfunc_(comm, BLKMPILSPAWN,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
		
		msg[0] = cid; msg[1] = numkids;
	}
/*
 * Broadcast the context ID for the parent-child intercommunicator and the
 * number of children spawned to the parents. In the case of an error
 * in spawning we are actually broadcasting the error code.
 */
	err = MPI_Bcast(msg, 2, MPI_INT, root, comm);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN, err));
	}
	
	if (rank != root) {

		cid = msg[0];
/*
 * A context ID of -1 means an error occurred in spawning so we
 * return with the error.
 */
		if (cid == -1) {
			err = msg[1];
			return(__lam_errfunc_(comm, BLKMPILSPAWN,
					__lam_mkerr_(MPI_ERR_OTHER, err)));
		}
/*
 * Allocate buffer to receive array of child GPS.
 */
		numkids = msg[1];
		kids = (struct _gps *) malloc((unsigned) (numkids *
				sizeof(struct _gps)));
		if (kids == 0) {
			return(__lam_errfunc_(comm, BLKMPILSPAWN,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
	}
/*
 * Broadcast the array of child GPS to parent group.
 */
	err = MPI_Bcast(kids, numkids * sizeof(struct _gps) / sizeof(int),
			MPI_INT, root, comm);
	if (err != MPI_SUCCESS) {
		free((char *) kids);
		return(__lam_errfunc_(comm, BLKMPILSPAWN, err));
	}
/*
 * Create the child group.
 */
	cg = (MPI_Group) malloc((unsigned) sizeof(struct _group) +
					(numkids * sizeof(struct _proc *)));
	if (cg == 0) {
		free((char *) kids);
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
	cg->g_nprocs = numkids;
	cg->g_myrank = MPI_UNDEFINED;
	cg->g_refcount = 1;
	cg->g_procs = (struct _proc **)
				((char *) cg + sizeof(struct _group));

	p = kids;
	pp = cg->g_procs;

	for (i = 0; i < numkids; ++i, ++p, ++pp) {

		*pp = __lam_procadd_(p);
		if (*pp == 0) {
			free((char *) kids);
			free((char *) cg);
			return(__lam_errfunc_(comm, BLKMPILSPAWN,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
	}
/*
 * Create the parent-child intercommunicator.
 */
	icomm = (MPI_Comm) malloc(sizeof(struct _comm));
	if (icomm == 0) {
		free((char *) kids);
		free((char *) cg);
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	*newcomm = icomm;
	icomm->c_inter = 1;
	icomm->c_refcount = 1;
	icomm->c_cube_dim = comm->c_cube_dim;
	icomm->c_group = comm->c_group;
	++(comm->c_group->g_refcount);
	icomm->c_rgroup = cg;
	icomm->c_contextid = cid;
	icomm->c_errhdl = comm->c_errhdl;
	++(comm->c_errhdl->eh_refcount);
	icomm->c_keys = 0;
	icomm->c_topo_type = MPI_UNDEFINED;
	icomm->c_topo_nprocs = 0;
	icomm->c_topo_ndims = 0;
	icomm->c_topo_dims = 0;
	icomm->c_topo_index = 0;
	icomm->c_topo_nedges = 0;
	icomm->c_topo_edges = 0;
	icomm->c_status = 0;

	if (! al_insert(__lam_comms_, &icomm)) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	if (__lam_trcomm_(icomm)) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_INTERN, errno)));
	}

	__lam_setcid_(cid);
/*
 * setup new processes
 */
	if (RPI_SPLIT(_rpi_lamd_addprocs, _rpi_c2c_addprocs, ())) {
		return(__lam_errfunc_(comm, BLKMPILSPAWN,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}
/*
 * Wait until all the children have initialized.
 * The root waits for rank 0 in the child world to communicate this fact and
 * then broadcasts it to the other parents.
 */
	if (rank == root) {
		err = MPI_Recv((void *) 0, 0, MPI_BYTE, 0, 0, icomm, &stat);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPILSPAWN, err));
		}
	}

	err = MPI_Bcast((void *) 0, 0, MPI_BYTE, root, comm);
	
	_LAM_TRACE(__lam_trcffend_(BLKMPILSPAWN, root, comm, 0, 0));
	
	__lam_resetfunc_(BLKMPILSPAWN);
	return(MPI_SUCCESS);
}

/*
 *	MPIL_Comm_parent
 *
 *	Function:	- returns the parent inter-communicator
 *	Accepts:	- communicator (out)
 */
int
MPIL_Comm_parent(comm)

MPI_Comm 		*comm;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPILCOMMPARENT);
/*
 * Check the arguments.
 */
	if (comm == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILCOMMPARENT,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set the intercomm.
 */
	*comm = lam_mpi_comm_parent;
	
	__lam_resetfunc_(BLKMPILCOMMPARENT);
	return(MPI_SUCCESS);
}

/*
 *	MPIL_Universe_size
 *
 *	Function:	- returns the number of nodes
 *	Accepts:	- number of nodes (out)
 */
int
MPIL_Universe_size(size)

int			*size;

{
	__lam_initerr_();
	__lam_setfunc_(BLKMPILUNIVERSESIZE);
/*
 * Check the arguments.
 */
	if (size == 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILUNIVERSESIZE,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set the intercomm.
 */
	*size = getncomp();

	if (*size < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPILUNIVERSESIZE,
				__lam_mkerr_(MPI_ERR_ARG, errno)));
	}
	
	__lam_resetfunc_(BLKMPILUNIVERSESIZE);
	return(MPI_SUCCESS);
}

/*
 *	spawn
 *
 *	Function:	- spawn MPI processes according to schema
 *	Accepts:	- schema
 *			- parent communicator
 *			- context ID for parent-child intercommunicator
 *			- ptr number of children (returned)
 *			- ptr array of child GPS (returned)
 *	Returns:	- 0 or LAMERROR
 */
static int
spawn(schema, comm, cid, numkids, kids)

char			*schema;
MPI_Comm		comm;
int			cid;
int			*numkids;
struct _gps		**kids;

{
	char		*appschema;		/* application schema */
	LIST		*appd;			/* application descriptor */
	int		junk;			/* junk argument */
	int		i;			/* favourite index*/
	int		nparent;		/* size of parent world */
	int		nworld;			/* size of child world */
	int4		rtf;			/* child runtime flags */
	struct nmsg	nhead;			/* network msg header */
	struct appproc	*pp;			/* ptr appl. process entry */
	struct _proc	**ppg;			/* ptr process in group */
	struct _gps	*procgps;		/* procs. GPS */
	struct _gps	*p;			/* favourite pointer */

/*
 * If schema is a single argument then it is an application schema file,
 * otherwise it is an explicit specification of what to execute like that
 * given to mpirun.
 */
	if (strchr(schema, ' ')) {
		appd = app_bufparse(schema, strlen(schema), &junk);
	} else {
		appschema = app_locate(schema);

		if (appschema == 0) {
			errno = EINVAL;
			return(LAMERROR);
		}

		appd = app_parse(appschema, &junk);
	}
	
	if (appd == 0) {
		errno = EUSAGE;
		return(LAMERROR);
	}
/*
 * Set environment inherited by children. The world spawning them consists
 * solely of the parent group.
 */
	rtf = _kio.ki_rtf;
	rtf &= ~(RTF_MPIRUN | RTF_TRON);
	MPI_Comm_size(comm, &nparent);
/*
 * Run the application.
 */
	if (app_run(appd, nparent, rtf, 0, 0)) return(LAMERROR);
/*
 * Send process table to all child processes.
 */
	nworld = al_count(appd);

	procgps = (struct _gps *)
		malloc((unsigned) (nworld + nparent) * sizeof(struct _gps));
	if (procgps == 0) return(LAMERROR);

	if (app_getprocs(appd, procgps)) {
		free((char *) procgps);
		return(LAMERROR);
	}
/*
 * Fill in child ranks in their MPI_COMM_WORLD.
 */
	for (i = 0, p = procgps; i < nworld; ++i, ++p) {
		p->gps_grank = i;
	}
/*
 * Fill in the parent world GPS.
 */
	ppg = comm->c_group->g_procs;

	for (i = 0; i < nparent; ++i, ++p, ++ppg) {
		*p = (*ppg)->p_gps;
	}
/*
 * Set up the message.
 */
	nhead.nh_type = 0;
	nhead.nh_flags = DINT4MSG;
	nhead.nh_msg = (char *) procgps;
	nhead.nh_length = (nworld + nparent) * sizeof(struct _gps);
	nhead.nh_data[1] = (int4) cid;
/*
 * Loop sending to each process.
 */
	for (pp = (struct appproc *) al_top(appd); pp;
			pp = (struct appproc *) al_next(appd, pp)) {

		nhead.nh_node = pp->app_node;
		nhead.nh_event = - pp->app_pid;

		if (nsend(&nhead)) {
			free((char *) procgps);
			return(LAMERROR);
		}
	}

	*numkids = nworld;
	*kids = procgps;

	return(0);
}
