/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD
 *
 *	$Log:	reduce.c,v $
 * Revision 6.0  96/02/29  14:13:27  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:25:17  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:16  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:04:56  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:36  gdburns
 * Ohio Release
 * 
 *	Function:	- reduce values to root
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- count of elements
 *			- type of elements
 *			- operation to perform
 *			- which process gets the result
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

/*
 * external functions
 */
extern void		__lam_dtcpy_();

/*
 * local functions
 */
static int		reduce_lin();
static int		reduce_log();

int
MPI_Reduce(sbuf, rbuf, count, dtype, op, root, comm)

void			*sbuf;
void			*rbuf;
int			count;
MPI_Datatype		dtype;
MPI_Op			op;
int			root;
MPI_Comm		comm;

{
	int		size;			/* group size */
	int		err;			/* error code */
	struct _gps	*p;			/* favourite pointer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIREDUCE);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return(__lam_errfunc_(comm, BLKMPIREDUCE,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm, BLKMPIREDUCE, 
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	MPI_Comm_size(comm, &size);

	if ((root < 0) || (root >= size)) {
		return(__lam_errfunc_(comm, BLKMPIREDUCE,
				__lam_mkerr_(MPI_ERR_ROOT, 0)));
	}

	if (op == MPI_OP_NULL) {
		return(__lam_errfunc_(comm, BLKMPIREDUCE,
				__lam_mkerr_(MPI_ERR_OP, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIREDUCE);
/*
 * Remember required parameters.
 */
	p = &(comm->c_group->g_procs[root]->p_gps);

	__lam_setparam_(BLKMPIREDUCE, root | (p->gps_grank << 16),
				(p->gps_node << 16) | p->gps_idx);
/*
 * Decide which algorithm to use.
 */
	if (size <= LAM_COLLMAXLIN) {
		err = reduce_lin(sbuf, rbuf, count, dtype, op, root, comm);
	} else {
		err = reduce_log(sbuf, rbuf, count, dtype, op, root, comm);
	}

	return(err);
}

/*
 *	reduce_lin
 *
 *	Function:	- reduction using O(N) algorithm
 *	Accepts:	- same as MPI_Reduce()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
reduce_lin(sbuf, rbuf, count, dtype, op, root, comm)

void			*sbuf;
void			*rbuf;
int			count;
MPI_Datatype		dtype;
MPI_Op			op;
int			root;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		err;			/* error code */
	char		*inmsg;			/* incoming message */
	char		*inbuf;			/* ptr incoming message */
	MPI_Status	status;			/* receive status */

	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &size);

	__lam_mkcoll_(comm);
/*
 * If not root, send data to the root.
 */
	if (rank != root) {
		err = MPI_Send(sbuf, count, dtype, root, BLKMPIREDUCE, comm);
		__lam_mkpt_(comm);

		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(comm, BLKMPIREDUCE, err));
		}
/*
 * Generate run time trace.
 */
		if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
			__lam_trcffend_(BLKMPIREDUCE,
						root, comm, dtype, count);
		}
		
		__lam_resetfunc_(BLKMPIREDUCE);
		return(MPI_SUCCESS);
	}
/*
 * Root receives and reduces messages.
 * Allocate buffer to receive messages.
 */
	if (size > 1) {
		inmsg = malloc((unsigned) count * dtype->dt_usrsize);
		if (inmsg == 0) {
			__lam_mkpt_(comm);
			return(__lam_errfunc_(comm, BLKMPIREDUCE,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}
	}
	else {
		inmsg = 0;
	}
/*
 * Initialize the receive buffer.
 */
	if (rank == (size - 1)) {
		err = __lam_dtsndrcv_(sbuf, count, dtype, rbuf, count,
					dtype, BLKMPIREDUCE, comm);
	} else {
		err = MPI_Recv(rbuf, count, dtype, size - 1,
					BLKMPIREDUCE, comm, &status);
	}

	if (err != MPI_SUCCESS) {
		if (inmsg) free(inmsg);
		__lam_mkpt_(comm);
		return(__lam_errfunc_(comm, BLKMPIREDUCE, err));
	}
/*
 * Loop receiving and calling reduction function (C or Fortran).
 */
	for (i = size - 2; i >= 0; --i) {

		if (rank == i) {
			inbuf = sbuf;
		}
		else {
			err = MPI_Recv(inmsg, count, dtype, i,
					BLKMPIREDUCE, comm, &status);

			if (err != MPI_SUCCESS) {
				free(inmsg);
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIREDUCE, err));
			}

			inbuf = inmsg;
		}
/*
 * Call reduction function.
 */
		if (op->op_f77dtype >= 0) {
			(op->op_func)(inbuf, rbuf, &count, &(op->op_f77dtype));
		} else {
			(op->op_func)(inbuf, rbuf, &count, &dtype);
		}				
	}

	if (inmsg) free(inmsg);
	__lam_mkpt_(comm);
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIREDUCE, root, comm, dtype, count);

	__lam_resetfunc_(BLKMPIREDUCE);
	return(MPI_SUCCESS);
}

/*
 *	reduce_log
 *
 *	Function:	- reduction using O(log N) algorithm
 *	Accepts:	- same as MPI_Reduce()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
reduce_log(sbuf, rbuf, count, dtype, op, root, comm)

void			*sbuf;
void			*rbuf;
int			count;
MPI_Datatype		dtype;
MPI_Op			op;
int			root;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		vrank;			/* virtual rank */
	int		err;			/* error code */
	int		peer;			/* peer rank */
	int		dim;			/* cube dimension */
	int		mask;			/* rank bit mask */
	int		fl_recv;		/* received msg flag */
	char		*buf1;			/* 1st temp buffer */
	char		*buf2;			/* 2nd temp buffer */
	void		*inmsg;			/* ptr incoming msg */
	void		*resmsg;		/* ptr result msg */
	MPI_Status	status;			/* receive status */
/*
 * Allocate the incoming and resulting message buffers.
 */
	buf1 = malloc((unsigned) count * dtype->dt_usrsize);
	buf2 = malloc((unsigned) count * dtype->dt_usrsize);

	if ((buf1 == 0) || (buf2 == 0)) {
		if (buf1) free(buf1);
		if (buf2) free(buf2);
		return(__lam_errfunc_(comm, BLKMPIREDUCE,
				__lam_mkerr_(MPI_ERR_OTHER, errno)));
	}

	MPI_Comm_size(comm, &size);
	MPI_Comm_rank(comm, &rank);

	vrank = (op->op_commute) ? (rank - root + size) % size : rank;

	__lam_mkcoll_(comm);

	dim = comm->c_cube_dim;
/*
 * Loop over cube dimensions.
 * High processes send to low ones in the dimension.
 */
	inmsg = buf1;
	resmsg = buf2;
	fl_recv = 0;

	for (i = 0, mask = 1; i < dim; ++i, mask <<= 1) {
/*
 * A high-proc sends to low-proc and stops.
 */
		if (vrank & mask) {

			peer = vrank & ~mask;
			if (op->op_commute) peer = (peer + root) % size;

			err = MPI_Send((fl_recv) ? resmsg : sbuf, count,
					dtype, peer, BLKMPIREDUCE, comm);
			if (err != MPI_SUCCESS) {
				free(buf1); free(buf2);
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIREDUCE, err));
			}

			break;
		}
/*
 * A low-proc receives, reduces, and moves to a higher dimension.
 */
		else {
			peer = vrank | mask;
			if (peer >= size) continue;
			if (op->op_commute) peer = (peer + root) % size;

			fl_recv = 1;

			err = MPI_Recv(inmsg, count, dtype, peer,
						BLKMPIREDUCE, comm, &status);
			if (err != MPI_SUCCESS) {
				free(buf1); free(buf2);
				__lam_mkpt_(comm);
				return(__lam_errfunc_(comm,
						BLKMPIREDUCE, err));
			}

			if (op->op_f77dtype >= 0) {
				(*op->op_func)((i > 0) ? resmsg : sbuf,
					inmsg, &count, &(op->op_f77dtype));
			} else {
				(*op->op_func)((i > 0) ? resmsg : sbuf,
					inmsg, &count, &dtype);
			}

			if (inmsg == buf1) {
				resmsg = buf1;
				inmsg = buf2;
			} else {
				resmsg = buf2;
				inmsg = buf1;
			}
		}
	}
/*
 * Get the result to the root if needed.
 */
	err = MPI_SUCCESS;

	if (vrank == 0) {
		if (root == rank) {
			__lam_dtcpy_(rbuf, (i > 0) ? resmsg : sbuf,
						count, dtype);
		} else {
			err = MPI_Send((i > 0) ? resmsg : sbuf, count,
					dtype, root, BLKMPIREDUCE, comm);
		}
	}
	else if (rank == root) {
		err = MPI_Recv(rbuf, count, dtype,
				0, BLKMPIREDUCE, comm, &status);
	}

	free(buf1);
	free(buf2);
	__lam_mkpt_(comm);

	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIREDUCE, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffend_(BLKMPIREDUCE, root, comm, dtype, count);

	__lam_resetfunc_(BLKMPIREDUCE);
	return(MPI_SUCCESS);
}
