/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/JRV
 *
 *	$Log:	reducescatter.c,v $
 * Revision 6.0  96/02/29  14:13:28  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/18  21:25:19  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/16  15:41:17  raja
 * Do not check for NULL buffers.
 * 
 * Revision 5.2  94/08/22  14:04:57  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:51:37  gdburns
 * Ohio Release
 * 
 *	Function:	- reduce and scatter data to all processes
 *	Accepts:	- send buffer
 *			- receive buffer
 *			- counts of elements to recv
 *			- type of elements
 *			- operation to perform
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <stdlib.h>

#include <blktype.h>
#include <mpi.h>
#include <terror.h>

int
MPI_Reduce_scatter(sbuf, rbuf, rcounts, dtype, op, comm)

void			*sbuf;
void			*rbuf;
int			*rcounts;
MPI_Datatype		dtype;
MPI_Op			op;
MPI_Comm		comm;

{
	int		i;			/* favourite index */
	int		size;			/* group size */
	int		rank;			/* my rank */
	int		count;			/* total count */
	int		err;			/* error code */
	int		*disps = 0;		/* displacement array */
	char		*bigbuf = 0;		/* reduce buffer */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIREDUCESCATTER);
/*
 * Check for invalid arguments.
 */
	if ((comm == MPI_COMM_NULL) || (comm->c_inter)) {
		return (__lam_errfunc_(comm, BLKMPIREDUCESCATTER,
				__lam_mkerr_(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL) {
		return(__lam_errfunc_(comm, BLKMPIREDUCESCATTER,
				__lam_mkerr_(MPI_ERR_TYPE, 0)));
	}

	if (op == MPI_OP_NULL) {
		return(__lam_errfunc_(comm, BLKMPIREDUCESCATTER,
				__lam_mkerr_(MPI_ERR_OP, 0)));
	}

	if (rcounts == 0) {
		return (__lam_errfunc_(comm, BLKMPIREDUCESCATTER,
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}
/*
 * Set up tracing.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON)
		__lam_trcffstart_(BLKMPIREDUCESCATTER);
/*
 * Initialize reduce & scatterv info at the root (rank 0).
 */
	MPI_Comm_size(comm, &size);
	MPI_Comm_rank(comm, &rank);

	for (i = 0, count = 0; i < size; ++i) count += rcounts[i];

	if (rank == 0) {
		disps = (int *) malloc((unsigned) size * sizeof(int));
		bigbuf = (void *) malloc((unsigned) count * dtype->dt_usrsize);

		if ((disps == 0) || (bigbuf == 0)) {
			if (disps) free((char *) disps);
			if (bigbuf) free(bigbuf);
			return (__lam_errfunc_(comm, BLKMPIREDUCESCATTER,
					__lam_mkerr_(MPI_ERR_OTHER, errno)));
		}

		disps[0] = 0;
		for (i = 0; i < (size - 1); ++i) {
			disps[i + 1] = disps[i] + rcounts[i];
		}
	}
/*
 * reduction
 */
	err = MPI_Reduce(sbuf, bigbuf, count, dtype, op, 0, comm);
	if (err != MPI_SUCCESS) {
		if (disps) free((char *) disps);
		if (bigbuf) free(bigbuf);
		return (__lam_errfunc_(comm, BLKMPIREDUCESCATTER, err));
	}
/*
 * scatter
 */
	err = MPI_Scatterv(bigbuf, rcounts, disps, dtype,
				rbuf, rcounts[rank], dtype, 0, comm);

	if (disps) free((char *) disps);
	if (bigbuf) free(bigbuf);

	if (err != MPI_SUCCESS) {
		return (__lam_errfunc_(comm, BLKMPIREDUCESCATTER, err));
	}
/*
 * Generate run time trace.
 */
	if ((_kio.ki_rtf & RTF_TRON) == RTF_TRON) {
		__lam_trcffend_(BLKMPIREDUCESCATTER,
					-1, comm, dtype, rcounts[rank]);
	}
	
	__lam_resetfunc_(BLKMPIREDUCESCATTER);
	return(MPI_SUCCESS);
}
