/*
 *	Ohio Trollius
 *	Copyright 1995 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	waitany.c,v $
 * Revision 6.0.1.1  96/05/10  15:12:50  nevin
 * Handle MPI_PROC_NULL
 * 
 * Revision 6.0  96/02/29  14:14:00  gdburns
 * Ohio Release
 * 
 * Revision 5.2.1.2  96/02/28  19:19:45  gdburns
 * major overhaul
 * 
 * Revision 5.2.1.1  94/09/26  14:07:54  raja
 * Nullify user's request.
 * 
 * Revision 5.2  94/08/22  14:05:23  gdburns
 * Ohio Release
 * 
 * Revision 5.1  94/05/18  12:52:07  gdburns
 * Ohio Release
 * 
 *	Function:	- wait for first request to finish
 *	Accepts:	- # of requests
 *			- array of requests
 *			- index of completed request (returned)
 *			- status (returned)
 *	Returns:	- MPI_SUCCESS or error code
 *			- in the case of an error return index is set to
 *			- MPI_UNDEFINED if the error is not associated with
 *			- the completed request, e.g. an internal error
 */

#include <blktype.h>
#include <mpi.h>
#include <mpitrace.h>

/*
 * external functions
 */
extern void		__lam_emptystat_();
extern void		__lam_initerr_();
extern void		__lam_resetfunc_();
extern void		__lam_setfunc_();
extern void		__lam_trmsg_();
extern void		_mpi_req_blkclr();
extern void		_mpi_req_blkset();
extern int		__lam_trincff_();
extern int		__lam_errfunc_();
extern int		__lam_mkerr_();
extern int		_mpi_req_advance();
extern int		_mpi_req_destroy();
extern int		_mpi_req_end();


int
MPI_Waitany(count, reqs, index, stat)

int			count;
MPI_Request		*reqs;
int			*index;
MPI_Status		*stat;

{
	int		i;			/* favourite index */
	int		nactive;		/* # active requests */
	int		err;			/* error code */
	MPI_Comm	comm;			/* communicator */
	MPI_Request	req;			/* request */
	int		fl_trace;		/* do tracing? */
	double		startt;			/* start time */
	double		finisht;		/* finish time */

	__lam_initerr_();
	__lam_setfunc_(BLKMPIWAITANY);

	*index = MPI_UNDEFINED;
	
	if (count < 0) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITANY, 
				__lam_mkerr_(MPI_ERR_COUNT, 0)));
	}

	if ((count > 0) && (reqs == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITANY,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}

	if ((index == 0) || (stat == 0)) {
		return(__lam_errfunc_(MPI_COMM_WORLD, BLKMPIWAITANY,
				__lam_mkerr_(MPI_ERR_ARG, 0)));
	}
/*
 * Set up tracing.
 */
	fl_trace = (_kio.ki_rtf & RTF_TRON) == RTF_TRON && ! __lam_trincff_();

	if (fl_trace) {
		startt = MPI_Wtime();
		_kio.ki_blktime = 0.0;
	}
	
	__lam_emptystat_(stat);
/*
 * While no request is done.
 */
	while (1) {
/*
 * Loop over the requests.
 */
		_mpi_req_blkclr();

		for (i = 0, nactive = 0; i < count; ++i) {
			req = reqs[i];
/*
 * Skip inactive requests.
 */
			if ((req == MPI_REQUEST_NULL) ||
					(req->rq_state == LAM_RQSINIT))
					continue;

			if (req->rq_state == LAM_RQSDONE) break;
/*
 * It's an active request.
 */
			++nactive;
			_mpi_req_blkset(req);
		}
/*
 * We found a completed request.
 */
		if (i < count) {
			*index = i;
			break;
		}
/*
 * If no active requests were found, we're done.
 */
		else if (nactive == 0) {
			*index = MPI_UNDEFINED;
			__lam_resetfunc_(BLKMPIWAITANY);
			return(MPI_SUCCESS);
		}
/*
 * Otherwise block on any and loop again.
 */
		else {
			err = _mpi_req_advance();
			if (err != MPI_SUCCESS) {
				*index = MPI_UNDEFINED;
				return(__lam_errfunc_(MPI_COMM_WORLD,
						BLKMPIWAITANY, err));
			}
		}
	}
/*
 * Finish the completed request.
 */
	_mpi_req_rem(req);

	err = _mpi_req_end(req);
	if (err != MPI_SUCCESS) {
		return(__lam_errfunc_(MPI_COMM_WORLD,
			BLKMPIWAITANY, err));
	}

	*stat = req->rq_status;
	comm = req->rq_comm;
/*
 * Generate a run time trace except in the case of cancelled request or
 * a request where the peer is MPI_PROC_NULL.
 */
	if (fl_trace && (req->rq_rank != MPI_PROC_NULL)
			&& !(req->rq_flags & LAM_RQFCANCEL)) {
		finisht = MPI_Wtime();

		__lam_trmsg_(
			(req->rq_type == LAM_RQIRECV) ? TRTINPUT : TRTNOIO,
			startt, LAM_S2US(finisht - startt - _kio.ki_blktime),
			LAM_S2US(_kio.ki_blktime), req->rq_rank, req->rq_tag,
			req->rq_comm, req->rq_dtype, req->rq_count,
			stat->MPI_SOURCE, stat->MPI_TAG, req->rq_seq,
			req->rq_type);
	}
/*
 * Reset persistent requests.
 * Destroy requests that are not persistent.
 */
	if (req->rq_marks & LAM_RQFPERSIST) {
		req->rq_state = LAM_RQSINIT;
	} else {
		err = _mpi_req_destroy(&reqs[i]);
		if (err != MPI_SUCCESS) {
			return(__lam_errfunc_(MPI_COMM_WORLD,
					BLKMPIWAITANY, err));
		}
	}

	__lam_resetfunc_(BLKMPIWAITANY);

	if (stat->MPI_ERROR != MPI_SUCCESS) {
		return(__lam_errfunc_(comm, BLKMPIWAITANY, stat->MPI_ERROR));
	}

	return(MPI_SUCCESS);
}
